# Copyright (c) 2023, Felix Fontein <felix@fontein.de>
# Copyright (c) 2023, Léo El Amri (@lel-amri)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import json
import os
import re
import shutil
import tempfile
import traceback
import typing as t
from collections import namedtuple
from shlex import quote

from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.docker.plugins.module_utils._logfmt import (
    InvalidLogFmt as _InvalidLogFmt,
)
from ansible_collections.community.docker.plugins.module_utils._logfmt import (
    parse_line as _parse_logfmt_line,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DockerBaseClass,
)
from ansible_collections.community.docker.plugins.module_utils._version import (
    LooseVersion,
)

PYYAML_IMPORT_ERROR: None | str  # pylint: disable=invalid-name
try:
    import yaml

    try:
        # use C version if possible for speedup
        from yaml import CSafeDumper as _SafeDumper
    except ImportError:
        from yaml import SafeDumper as _SafeDumper  # type: ignore
except ImportError:
    HAS_PYYAML = False
    PYYAML_IMPORT_ERROR = traceback.format_exc()  # pylint: disable=invalid-name
else:
    HAS_PYYAML = True
    PYYAML_IMPORT_ERROR = None  # pylint: disable=invalid-name

if t.TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from ansible_collections.community.docker.plugins.module_utils._common_cli import (
        AnsibleModuleDockerClient as _Client,
    )


DOCKER_COMPOSE_FILES = (
    "compose.yaml",
    "compose.yml",
    "docker-compose.yaml",
    "docker-compose.yml",
)

DOCKER_STATUS_DONE = frozenset(
    (
        "Started",
        "Healthy",
        "Exited",
        "Restarted",
        "Running",
        "Created",
        "Stopped",
        "Killed",
        "Removed",
        # An extra, specific to containers
        "Recreated",
        # Extras for pull events
        "Pulled",
        # Extras for built events
        "Built",
    )
)
DOCKER_STATUS_WORKING = frozenset(
    (
        "Creating",
        "Starting",
        "Restarting",
        "Stopping",
        "Killing",
        "Removing",
        # An extra, specific to containers
        "Recreate",
        # Extras for pull events
        "Pulling",
        # Extras for build start events
        "Building",
    )
)
DOCKER_STATUS_PULL = frozenset(
    (
        "Pulled",
        "Pulling",
    )
)
DOCKER_STATUS_BUILD = frozenset(
    (
        "Built",
        "Building",
    )
)
DOCKER_STATUS_ERROR = frozenset(("Error",))
DOCKER_STATUS_WARNING = frozenset(("Warning",))
DOCKER_STATUS_WAITING = frozenset(("Waiting",))
DOCKER_STATUS = frozenset(
    DOCKER_STATUS_DONE
    | DOCKER_STATUS_WORKING
    | DOCKER_STATUS_PULL
    | DOCKER_STATUS_ERROR
    | DOCKER_STATUS_WAITING
)
DOCKER_STATUS_AND_WARNING = frozenset(DOCKER_STATUS | DOCKER_STATUS_WARNING)

DOCKER_PULL_PROGRESS_DONE = frozenset(
    (
        "Already exists",
        "Download complete",
        "Pull complete",
    )
)
DOCKER_PULL_PROGRESS_WORKING_OLD = frozenset(
    (
        "Pulling fs layer",
        "Waiting",
        "Downloading",
        "Verifying Checksum",
        "Extracting",
    )
)
DOCKER_PULL_PROGRESS_WORKING = frozenset(DOCKER_PULL_PROGRESS_WORKING_OLD | {"Working"})


class ResourceType:
    UNKNOWN = "unknown"
    NETWORK = "network"
    IMAGE = "image"
    IMAGE_LAYER = "image-layer"
    VOLUME = "volume"
    CONTAINER = "container"
    SERVICE = "service"

    @classmethod
    def from_docker_compose_event(cls, resource_type: str) -> t.Any:
        return {
            "Network": cls.NETWORK,
            "Image": cls.IMAGE,
            "Volume": cls.VOLUME,
            "Container": cls.CONTAINER,
            "Service": cls.SERVICE,
        }[resource_type]


Event = namedtuple("Event", ["resource_type", "resource_id", "status", "msg"])


_DRY_RUN_MARKER = "DRY-RUN MODE -"

_RE_RESOURCE_EVENT = re.compile(
    r"^"
    r"\s*"
    r"(?P<resource_type>Network|Image|Volume|Container)"
    r"\s+"
    r"(?P<resource_id>\S+)"
    r"\s+"
    r"(?P<status>\S(?:|.*\S))"
    r"\s*"
    r"$"
)

_RE_PULL_EVENT = re.compile(
    r"^"
    r"\s*"
    r"(?P<service>\S+)"
    r"\s+"
    f"(?P<status>{'|'.join(re.escape(status) for status in DOCKER_STATUS_PULL)})"
    r"\s*"
    r"$"
)

_DOCKER_PULL_PROGRESS_WD = sorted(
    DOCKER_PULL_PROGRESS_DONE | DOCKER_PULL_PROGRESS_WORKING_OLD
)

_RE_PULL_PROGRESS = re.compile(
    r"^"
    r"\s*"
    r"(?P<layer>\S+)"
    r"\s+"
    f"(?P<status>{'|'.join(re.escape(status) for status in _DOCKER_PULL_PROGRESS_WD)})"
    r"\s*"
    r"(?:|\s\[[^]]+\]\s+\S+\s*|\s+[0-9.kKmMgGbB]+/[0-9.kKmMgGbB]+\s*)"
    r"$"
)

_RE_ERROR_EVENT = re.compile(
    r"^"
    r"\s*"
    r"(?P<resource_id>\S+)"
    r"\s+"
    f"(?P<status>{'|'.join(re.escape(status) for status in DOCKER_STATUS_ERROR)})"
    r"\s*"
    r"(?P<msg>\S.*\S)?"
    r"$"
)

_RE_WARNING_EVENT = re.compile(
    r"^"
    r"\s*"
    r"(?P<resource_id>\S+)"
    r"\s+"
    f"(?P<status>{'|'.join(re.escape(status) for status in DOCKER_STATUS_WARNING)})"
    r"\s*"
    r"(?P<msg>\S.*\S)?"
    r"$"
)

_RE_CONTINUE_EVENT = re.compile(r"^\s*(?P<resource_id>\S+)\s+-\s*(?P<msg>\S(?:|.*\S))$")

_RE_SKIPPED_EVENT = re.compile(
    r"^"
    r"\s*"
    r"(?P<resource_id>\S+)"
    r"\s+"
    r"Skipped(?: -"
    r"\s*"
    r"(?P<msg>\S(?:|.*\S))|\s*)"
    r"$"
)

_RE_BUILD_START_EVENT = re.compile(r"^\s*build service\s+(?P<resource_id>\S+)$")

_RE_BUILD_PROGRESS_EVENT = re.compile(r"^\s*==>\s+(?P<msg>.*)$")

# The following needs to be kept in sync with the MINIMUM_VERSION compose_v2 docs fragment
MINIMUM_COMPOSE_VERSION = "2.18.0"


def _extract_event(
    line: str, warn_function: Callable[[str], None] | None = None
) -> tuple[Event | None, bool]:
    match = _RE_RESOURCE_EVENT.match(line)
    if match is not None:
        status = match.group("status")
        msg = None
        if status not in DOCKER_STATUS:
            status, msg = msg, status
        return (
            Event(
                ResourceType.from_docker_compose_event(match.group("resource_type")),
                match.group("resource_id"),
                status,
                msg,
            ),
            True,
        )
    match = _RE_PULL_EVENT.match(line)
    if match:
        return (
            Event(
                ResourceType.SERVICE,
                match.group("service"),
                match.group("status"),
                None,
            ),
            True,
        )
    match = _RE_ERROR_EVENT.match(line)
    if match:
        return (
            Event(
                ResourceType.UNKNOWN,
                match.group("resource_id"),
                match.group("status"),
                match.group("msg") or None,
            ),
            True,
        )
    match = _RE_WARNING_EVENT.match(line)
    if match:
        if warn_function:
            if match.group("msg"):
                msg = f"{match.group('resource_id')}: {match.group('msg')}"
            else:
                msg = f"Unspecified warning for {match.group('resource_id')}"
            warn_function(msg)
        return None, True
    match = _RE_PULL_PROGRESS.match(line)
    if match:
        return (
            Event(
                ResourceType.IMAGE_LAYER,
                match.group("layer"),
                match.group("status"),
                None,
            ),
            True,
        )
    match = _RE_SKIPPED_EVENT.match(line)
    if match:
        return (
            Event(
                ResourceType.UNKNOWN,
                match.group("resource_id"),
                "Skipped",
                match.group("msg"),
            ),
            True,
        )
    match = _RE_BUILD_START_EVENT.match(line)
    if match:
        return (
            Event(
                ResourceType.SERVICE,
                match.group("resource_id"),
                "Building",
                None,
            ),
            True,
        )
    return None, False


def _extract_logfmt_event(
    line: str, warn_function: Callable[[str], None] | None = None
) -> tuple[Event | None, bool]:
    try:
        result = _parse_logfmt_line(line, logrus_mode=True)
    except _InvalidLogFmt:
        return None, False
    if "time" not in result or "level" not in result or "msg" not in result:
        return None, False
    if result["level"] == "warning":
        if warn_function:
            warn_function(result["msg"])
        return None, True
    # TODO: no idea what to do with this
    return None, False


def _warn_missing_dry_run_prefix(
    line: str,
    warn_missing_dry_run_prefix: bool,
    warn_function: Callable[[str], None] | None,
) -> None:
    if warn_missing_dry_run_prefix and warn_function:
        # This could be a bug, a change of docker compose's output format, ...
        # Tell the user to report it to us :-)
        warn_function(
            f"Event line is missing dry-run mode marker: {line!r}. Please check with the latest community.docker version,"
            " and if the problem still happens there, please report this at "
            "https://github.com/ansible-collections/community.docker/issues/new?assignees=&labels=&projects=&template=bug_report.md"
        )


def _warn_unparsable_line(
    line: str, warn_function: Callable[[str], None] | None
) -> None:
    # This could be a bug, a change of docker compose's output format, ...
    # Tell the user to report it to us :-)
    if warn_function:
        warn_function(
            f"Cannot parse event from line: {line!r}. Please check with the latest community.docker version,"
            " and if the problem still happens there, please report this at "
            "https://github.com/ansible-collections/community.docker/issues/new?assignees=&labels=&projects=&template=bug_report.md"
        )


def _find_last_event_for(
    events: list[Event], resource_id: str
) -> tuple[int, Event] | None:
    for index, event in enumerate(reversed(events)):
        if event.resource_id == resource_id:
            return len(events) - 1 - index, event
    return None


def _concat_event_msg(event: Event, append_msg: str) -> Event:
    return Event(
        event.resource_type,
        event.resource_id,
        event.status,
        "\n".join(msg for msg in [event.msg, append_msg] if msg is not None),
    )


_JSON_LEVEL_TO_STATUS_MAP = {
    "warning": "Warning",
    "error": "Error",
}


def parse_json_events(
    stderr: bytes, warn_function: Callable[[str], None] | None = None
) -> list[Event]:
    events = []
    stderr_lines = stderr.splitlines()
    if stderr_lines and stderr_lines[-1] == b"":
        del stderr_lines[-1]
    for line in stderr_lines:
        line = line.strip()
        if not line.startswith(b"{") or not line.endswith(b"}"):
            if line.startswith(b"Warning: "):
                # This is a bug in Compose that will get fixed by https://github.com/docker/compose/pull/11996
                event = Event(
                    ResourceType.UNKNOWN,
                    None,
                    "Warning",
                    to_text(line[len(b"Warning: ") :]),
                )
                events.append(event)
                continue
            if warn_function:
                warn_function(
                    f"Cannot parse event from non-JSON line: {line!r}. Please check with the latest community.docker version,"
                    " and if the problem still happens there, please report this at "
                    "https://github.com/ansible-collections/community.docker/issues/new?assignees=&labels=&projects=&template=bug_report.md"
                )
            continue
        try:
            line_data = json.loads(line)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            if warn_function:
                warn_function(
                    f"Cannot parse event from line: {line!r}: {exc}. Please check with the latest community.docker version,"
                    " and if the problem still happens there, please report this at "
                    "https://github.com/ansible-collections/community.docker/issues/new?assignees=&labels=&projects=&template=bug_report.md"
                )
            continue
        if line_data.get("tail"):
            resource_type = ResourceType.UNKNOWN
            msg = line_data.get("text")
            status = "Error"
            if isinstance(msg, str) and msg.lower().startswith("warning:"):
                # For some reason, Writer.TailMsgf() is always used for errors *except* in one place,
                # where its message is prepended with 'WARNING: ' (in pkg/compose/pull.go).
                status = "Warning"
                msg = msg[len("warning:") :].lstrip()
            event = Event(
                resource_type,
                None,
                status,
                msg,
            )
        elif line_data.get("error"):
            resource_type = ResourceType.UNKNOWN
            event = Event(
                resource_type,
                line_data.get("id"),
                "Error",
                line_data.get("message"),
            )
        else:
            resource_type = ResourceType.UNKNOWN
            resource_id = line_data.get("id")
            status = line_data.get("status")
            text = line_data.get("text")
            if resource_id == " " and text and text.startswith("build service "):
                # Example:
                # {"dry-run":true,"id":" ","text":"build service app"}
                resource_id = "S" + text[len("build s") :]
                text = "Building"
            if (
                isinstance(resource_id, str)
                and resource_id.endswith("==>")
                and text
                and text.startswith("==> writing image ")
            ):
                # Example:
                # {"dry-run":true,"id":"==>","text":"==> writing image dryRun-7d1043473d55bfa90e8530d35801d4e381bc69f0"}
                # {"dry-run":true,"id":"ansible-docker-test-dc713f1f-container ==>","text":"==> writing image dryRun-5d9204268db1a73d57bbd24a25afbeacebe2bc02"}
                # (The longer form happens since Docker Compose 2.39.0)
                continue
            if (
                isinstance(resource_id, str)
                and resource_id.endswith("==> ==>")
                and text
                and text.startswith("naming to ")
            ):
                # Example:
                # {"dry-run":true,"id":"==> ==>","text":"naming to display-app"}
                # {"dry-run":true,"id":"ansible-docker-test-dc713f1f-container ==> ==>","text":"naming to ansible-docker-test-dc713f1f-image"}
                # (The longer form happens since Docker Compose 2.39.0)
                continue
            if (
                status in ("Working", "Done")
                and isinstance(line_data.get("parent_id"), str)
                and line_data["parent_id"].startswith("Image ")
            ):
                # Compose 5.0.0+:
                # {"id":"63a26ae4e8a8","parent_id":"Image ghcr.io/ansible-collections/simple-1:tag","status":"Working"}
                # {"id":"63a26ae4e8a8","parent_id":"Image ghcr.io/ansible-collections/simple-1:tag","status":"Done","percent":100}
                resource_type = ResourceType.IMAGE_LAYER
                resource_id = line_data["parent_id"][len("Image ") :]
            elif isinstance(resource_id, str) and " " in resource_id:
                resource_type_str, resource_id = resource_id.split(" ", 1)
                try:
                    resource_type = ResourceType.from_docker_compose_event(
                        resource_type_str
                    )
                except KeyError:
                    if warn_function:
                        warn_function(
                            f"Unknown resource type {resource_type_str!r} in line {line!r}. Please check with the latest community.docker version,"
                            " and if the problem still happens there, please report this at "
                            "https://github.com/ansible-collections/community.docker/issues/new?assignees=&labels=&projects=&template=bug_report.md"
                        )
                    resource_type = ResourceType.UNKNOWN
            elif text in DOCKER_STATUS_PULL:
                resource_type = ResourceType.IMAGE
                status, text = text, status
            elif (
                text in DOCKER_PULL_PROGRESS_DONE
                or line_data.get("text") in DOCKER_PULL_PROGRESS_WORKING_OLD
            ):
                resource_type = ResourceType.IMAGE_LAYER
                status, text = text, status
            elif (
                status is None
                and isinstance(text, str)
                and text.startswith("Skipped - ")
            ):
                status, text = text.split(" - ", 1)
            elif (
                line_data.get("level") in _JSON_LEVEL_TO_STATUS_MAP
                and "msg" in line_data
            ):
                status = _JSON_LEVEL_TO_STATUS_MAP[line_data["level"]]
                text = line_data["msg"]
            if (
                status not in DOCKER_STATUS_AND_WARNING
                and text in DOCKER_STATUS_AND_WARNING
            ):
                status, text = text, status
            event = Event(
                resource_type,
                resource_id,
                status,
                text,
            )

        events.append(event)
    return events


def parse_events(
    stderr: bytes,
    dry_run: bool = False,
    warn_function: Callable[[str], None] | None = None,
    nonzero_rc: bool = False,
) -> list[Event]:
    events = []
    error_event = None
    stderr_lines = stderr.splitlines()
    if stderr_lines and stderr_lines[-1] == b"":
        del stderr_lines[-1]
    for index, line_b in enumerate(stderr_lines):
        line = to_text(line_b.strip())
        if not line:
            continue
        warn_missing_dry_run_prefix = False
        if dry_run:
            if line.startswith(_DRY_RUN_MARKER):
                line = line[len(_DRY_RUN_MARKER) :].lstrip()
            else:
                warn_missing_dry_run_prefix = True
        event, parsed = _extract_event(line, warn_function=warn_function)
        if event is not None:
            events.append(event)
            if event.status in DOCKER_STATUS_ERROR:
                error_event = event
            else:
                error_event = None
            _warn_missing_dry_run_prefix(
                line, warn_missing_dry_run_prefix, warn_function
            )
            continue
        if parsed:
            continue
        match = _RE_BUILD_PROGRESS_EVENT.match(line)
        if match:
            # Ignore this
            continue
        match = _RE_CONTINUE_EVENT.match(line)
        if match:
            # Continuing an existing event
            index_event = _find_last_event_for(events, match.group("resource_id"))
            if index_event is not None:
                index, event = index_event
                events[index] = _concat_event_msg(event, match.group("msg"))
        event, parsed = _extract_logfmt_event(line, warn_function=warn_function)
        if event is not None:
            events.append(event)
        elif parsed:
            continue
        if error_event is not None:
            # Unparsable line that apparently belongs to the previous error event
            events[-1] = _concat_event_msg(error_event, line)
            continue
        if line.startswith("Error "):
            # Error message that is independent of an error event
            error_event = Event(
                ResourceType.UNKNOWN,
                "",
                "Error",
                line,
            )
            events.append(error_event)
            continue
        if len(stderr_lines) == 1 or (nonzero_rc and index == len(stderr_lines) - 1):
            # **Very likely** an error message that is independent of an error event
            error_event = Event(
                ResourceType.UNKNOWN,
                "",
                "Error",
                line,
            )
            events.append(error_event)
            continue
        _warn_missing_dry_run_prefix(line, warn_missing_dry_run_prefix, warn_function)
        _warn_unparsable_line(line, warn_function)
    return events


def has_changes(
    events: Sequence[Event],
    ignore_service_pull_events: bool = False,
    ignore_build_events: bool = False,
) -> bool:
    for event in events:
        if event.status in DOCKER_STATUS_WORKING:
            if ignore_service_pull_events and event.status in DOCKER_STATUS_PULL:
                continue
            if ignore_build_events and event.status in DOCKER_STATUS_BUILD:
                continue
            return True
        if (
            event.resource_type == ResourceType.IMAGE_LAYER
            and event.status in DOCKER_PULL_PROGRESS_WORKING
        ):
            return True
    return False


def extract_actions(events: Sequence[Event]) -> list[dict[str, t.Any]]:
    actions = []
    pull_actions = set()
    for event in events:
        if (
            event.resource_type == ResourceType.IMAGE_LAYER
            and event.status in DOCKER_PULL_PROGRESS_WORKING
        ):
            pull_id = (event.resource_id, event.status)
            if pull_id not in pull_actions:
                pull_actions.add(pull_id)
                actions.append(
                    {
                        "what": event.resource_type,
                        "id": event.resource_id,
                        "status": event.status,
                    }
                )
        if (
            event.resource_type != ResourceType.IMAGE_LAYER
            and event.status in DOCKER_STATUS_WORKING
        ):
            actions.append(
                {
                    "what": event.resource_type,
                    "id": event.resource_id,
                    "status": event.status,
                }
            )
    return actions


def emit_warnings(
    events: Sequence[Event], warn_function: Callable[[str], None]
) -> None:
    for event in events:
        # If a message is present, assume it is a warning
        if (
            event.status is None and event.msg is not None
        ) or event.status in DOCKER_STATUS_WARNING:
            warn_function(
                f"Docker compose: {event.resource_type} {event.resource_id}: {event.msg}"
            )


def is_failed(events: Sequence[Event], rc: int) -> bool:
    return bool(rc)


def update_failed(
    result: dict[str, t.Any],
    events: Sequence[Event],
    args: list[str],
    stdout: str | bytes,
    stderr: str | bytes,
    rc: int,
    cli: str,
) -> bool:
    if not rc:
        return False
    errors = []
    for event in events:
        if event.status in DOCKER_STATUS_ERROR:
            if event.resource_id is None:
                if event.resource_type == "unknown":
                    msg = (
                        "General error: "
                        if event.resource_type == "unknown"
                        else f"Error when processing {event.resource_type}: "
                    )
            else:
                msg = (
                    f"Error when processing {event.resource_type} {event.resource_id}: "
                )
                if event.resource_type == "unknown":
                    msg = f"Error when processing {event.resource_id}: "
                    if event.resource_id == "":
                        msg = "General error: "
            msg += f"{event.status}" if event.msg is None else f"{event.msg}"
            errors.append(msg)
    if not errors:
        errors.append(f"Return code {rc} is non-zero")
    result["failed"] = True
    result["msg"] = "\n".join(errors)
    result["cmd"] = " ".join(quote(arg) for arg in [cli] + args)
    result["stdout"] = to_text(stdout)
    result["stderr"] = to_text(stderr)
    result["rc"] = rc
    return True


def common_compose_argspec() -> dict[str, t.Any]:
    return {
        "project_src": {"type": "path"},
        "project_name": {"type": "str"},
        "files": {"type": "list", "elements": "path"},
        "definition": {"type": "dict"},
        "env_files": {"type": "list", "elements": "path"},
        "profiles": {"type": "list", "elements": "str"},
        "check_files_existing": {"type": "bool", "default": True},
    }


def common_compose_argspec_ex() -> dict[str, t.Any]:
    return {
        "argspec": common_compose_argspec(),
        "mutually_exclusive": [("definition", "project_src"), ("definition", "files")],
        "required_one_of": [
            ("definition", "project_src"),
        ],
        "required_by": {
            "definition": ("project_name",),
        },
    }


def combine_binary_output(*outputs: bytes | None) -> bytes:
    return b"\n".join(out for out in outputs if out)


def combine_text_output(*outputs: str | None) -> str:
    return "\n".join(out for out in outputs if out)


class BaseComposeManager(DockerBaseClass):
    def __init__(
        self, client: _Client, min_version: str = MINIMUM_COMPOSE_VERSION
    ) -> None:
        super().__init__()
        self.client = client
        self.check_mode = self.client.check_mode
        self.cleanup_dirs = set()
        parameters = self.client.module.params

        if parameters["definition"] is not None and not HAS_PYYAML:
            self.fail(missing_required_lib("PyYAML"), exception=PYYAML_IMPORT_ERROR)

        self.project_name = parameters["project_name"]
        if parameters["definition"] is not None:
            self.project_src = tempfile.mkdtemp(prefix="ansible")
            self.cleanup_dirs.add(self.project_src)
            compose_file = os.path.join(self.project_src, "compose.yaml")
            self.client.module.add_cleanup_file(compose_file)
            try:
                with open(compose_file, "wb") as f:
                    yaml.dump(
                        parameters["definition"],
                        f,
                        encoding="utf-8",
                        Dumper=_SafeDumper,
                    )
            except Exception as exc:  # pylint: disable=broad-exception-caught
                self.fail(f"Error writing to {compose_file} - {exc}")
        else:
            self.project_src = os.path.abspath(parameters["project_src"])

        self.files = parameters["files"]
        self.env_files = parameters["env_files"]
        self.profiles = parameters["profiles"]

        compose_version = self.get_compose_version()
        self.compose_version = LooseVersion(compose_version)
        if self.compose_version < LooseVersion(min_version):
            self.fail(
                f"Docker CLI {self.client.get_cli()} has the compose plugin with version {compose_version}; need version {min_version} or later"
            )

        if not os.path.isdir(self.project_src):
            self.fail(f'"{self.project_src}" is not a directory')

        self.check_files_existing = parameters["check_files_existing"]
        if self.files:
            for file in self.files:
                path = os.path.join(self.project_src, file)
                if not os.path.exists(path):
                    self.fail(
                        f'Cannot find Compose file "{file}" relative to project directory "{self.project_src}"'
                    )
        elif self.check_files_existing and all(
            not os.path.exists(os.path.join(self.project_src, f))
            for f in DOCKER_COMPOSE_FILES
        ):
            filenames = ", ".join(DOCKER_COMPOSE_FILES[:-1])
            self.fail(
                f'"{self.project_src}" does not contain {filenames}, or {DOCKER_COMPOSE_FILES[-1]}'
            )

        # Support for JSON output was added in Compose 2.29.0 (https://github.com/docker/compose/releases/tag/v2.29.0);
        # more precisely in https://github.com/docker/compose/pull/11478
        self.use_json_events = self.compose_version >= LooseVersion("2.29.0")

    def get_compose_version(self) -> str:
        return (
            self.get_compose_version_from_cli() or self.get_compose_version_from_api()
        )

    def get_compose_version_from_cli(self) -> str | None:
        rc, version_info, dummy_stderr = self.client.call_cli(
            "compose", "version", "--format", "json"
        )
        if rc:
            return None
        try:
            version = json.loads(version_info)["version"]
            if version == "dev":
                return None
            return version.lstrip("v")
        except Exception:  # pylint: disable=broad-exception-caught
            return None

    def get_compose_version_from_api(self) -> str:
        compose = self.client.get_client_plugin_info("compose")
        if compose is None:
            self.fail(
                f"Docker CLI {self.client.get_cli()} does not have the compose plugin installed"
            )
        if compose["Version"] == "dev":
            self.fail(
                f'Docker CLI {self.client.get_cli()} has a compose plugin installed, but it reports version "dev".'
                " Please use a version of the plugin that returns a proper version."
            )
        return compose["Version"].lstrip("v")

    def fail(self, msg: str, **kwargs: t.Any) -> t.NoReturn:
        self.cleanup()
        self.client.fail(msg, **kwargs)

    def get_base_args(self, plain_progress: bool = False) -> list[str]:
        args = ["compose", "--ansi", "never"]
        if self.use_json_events and not plain_progress:
            args.extend(["--progress", "json"])
        elif self.compose_version >= LooseVersion("2.19.0"):
            # https://github.com/docker/compose/pull/10690
            args.extend(["--progress", "plain"])
        args.extend(["--project-directory", self.project_src])
        if self.project_name:
            args.extend(["--project-name", self.project_name])
        for file in self.files or []:
            args.extend(["--file", file])
        for env_file in self.env_files or []:
            args.extend(["--env-file", env_file])
        for profile in self.profiles or []:
            args.extend(["--profile", profile])
        return args

    def _handle_failed_cli_call(
        self, args: list[str], rc: int, stdout: str | bytes, stderr: bytes
    ) -> t.NoReturn:
        events = parse_json_events(stderr, warn_function=self.client.warn)
        result: dict[str, t.Any] = {}
        self.update_failed(result, events, args, stdout, stderr, rc)
        self.client.module.exit_json(**result)

    def list_containers_raw(self) -> list[dict[str, t.Any]]:
        args = self.get_base_args() + ["ps", "--format", "json", "--all"]
        if self.compose_version >= LooseVersion("2.23.0"):
            # https://github.com/docker/compose/pull/11038
            args.append("--no-trunc")
        if self.compose_version >= LooseVersion("2.21.0"):
            # Breaking change in 2.21.0: https://github.com/docker/compose/pull/10918
            rc, containers, stderr = self.client.call_cli_json_stream(
                *args, cwd=self.project_src, check_rc=not self.use_json_events
            )
        else:
            rc, containers, stderr = self.client.call_cli_json(
                *args, cwd=self.project_src, check_rc=not self.use_json_events
            )
        if self.use_json_events and rc != 0:
            self._handle_failed_cli_call(args, rc, json.dumps(containers), stderr)
        return containers

    def list_containers(self) -> list[dict[str, t.Any]]:
        result = []
        for container in self.list_containers_raw():
            labels = {}
            if container.get("Labels"):
                for part in container["Labels"].split(","):
                    label_value = part.split("=", 1)
                    labels[label_value[0]] = (
                        label_value[1] if len(label_value) > 1 else ""
                    )
            container["Labels"] = labels
            container["Names"] = container.get("Names", container["Name"]).split(",")
            container["Networks"] = container.get("Networks", "").split(",")
            container["Publishers"] = container.get("Publishers") or []
            result.append(container)
        return result

    def list_images(self) -> list[str]:
        args = self.get_base_args() + ["images", "--format", "json"]
        rc, images, stderr = self.client.call_cli_json(
            *args, cwd=self.project_src, check_rc=not self.use_json_events
        )
        if self.use_json_events and rc != 0:
            self._handle_failed_cli_call(args, rc, images, stderr)
        if isinstance(images, dict):
            # Handle breaking change in Docker Compose 2.37.0; see
            # https://github.com/ansible-collections/community.docker/issues/1082
            # and https://github.com/docker/compose/issues/12916 for details
            images = list(images.values())
        return images

    def parse_events(
        self, stderr: bytes, dry_run: bool = False, nonzero_rc: bool = False
    ) -> list[Event]:
        if self.use_json_events:
            return parse_json_events(stderr, warn_function=self.client.warn)
        return parse_events(
            stderr,
            dry_run=dry_run,
            warn_function=self.client.warn,
            nonzero_rc=nonzero_rc,
        )

    def emit_warnings(self, events: Sequence[Event]) -> None:
        emit_warnings(events, warn_function=self.client.warn)

    def update_result(
        self,
        result: dict[str, t.Any],
        events: Sequence[Event],
        stdout: str | bytes,
        stderr: str | bytes,
        ignore_service_pull_events: bool = False,
        ignore_build_events: bool = False,
    ) -> None:
        result["changed"] = result.get("changed", False) or has_changes(
            events,
            ignore_service_pull_events=ignore_service_pull_events,
            ignore_build_events=ignore_build_events,
        )
        result["actions"] = result.get("actions", []) + extract_actions(events)
        result["stdout"] = combine_text_output(result.get("stdout"), to_text(stdout))
        result["stderr"] = combine_text_output(result.get("stderr"), to_text(stderr))

    def update_failed(
        self,
        result: dict[str, t.Any],
        events: Sequence[Event],
        args: list[str],
        stdout: str | bytes,
        stderr: bytes,
        rc: int,
    ) -> bool:
        return update_failed(
            result,
            events,
            args=args,
            stdout=stdout,
            stderr=stderr,
            rc=rc,
            cli=self.client.get_cli(),
        )

    def cleanup_result(self, result: dict[str, t.Any]) -> None:
        if not result.get("failed"):
            # Only return stdout and stderr if it is not empty
            for res in ("stdout", "stderr"):
                if result.get(res) == "":
                    result.pop(res)

    def cleanup(self) -> None:
        for directory in self.cleanup_dirs:
            try:
                shutil.rmtree(directory, True)
            except Exception:  # pylint: disable=broad-exception-caught
                # should not happen, but simply ignore to be on the safe side
                pass



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: ecs_service
version_added: 1.0.0
short_description: Create, terminate, start or stop a service in ECS
description:
  - Creates or terminates ECS services.
notes:
  - The service role specified must be assumable. (i.e. have a trust relationship for the ecs service, ecs.amazonaws.com)
  - For details of the parameters and returns see U(https://boto3.readthedocs.io/en/latest/reference/services/ecs.html).
  - An IAM role must have been previously created.
author:
  - "Mark Chance (@Java1Guy)"
  - "Darek Kaczynski (@kaczynskid)"
  - "Stephane Maarek (@simplesteph)"
  - "Zac Blazic (@zacblazic)"
options:
    state:
        description:
          - The desired state of the service.
        required: true
        choices: ["present", "absent", "deleting"]
        type: str
    name:
        description:
          - The name of the service.
        required: true
        type: str
        aliases: ['service']
    cluster:
        description:
          - The name of the cluster in which the service exists.
          - If not specified, the cluster name will be C(default).
        required: false
        type: str
        default: 'default'
    task_definition:
        description:
          - The task definition the service will run.
          - This parameter is required when I(state=present) unless I(force_new_deployment=True).
          - This parameter is ignored when updating a service with a C(CODE_DEPLOY) deployment controller in which case
            the task definition is managed by Code Pipeline and cannot be updated.
        required: false
        type: str
    load_balancers:
        description:
          - The list of ELBs defined for this service.
          - Load balancers for an existing service cannot be updated, and it is an error to do so.
          - When the deployment controller is CODE_DEPLOY changes to this value are simply ignored, and do not cause an error.
        required: false
        type: list
        elements: dict
        default: []
    desired_count:
        description:
          - The count of how many instances of the service.
          - This parameter is required when I(state=present).
        required: false
        type: int
    client_token:
        description:
          - Unique, case-sensitive identifier you provide to ensure the idempotency of the request. Up to 32 ASCII characters are allowed.
        required: false
        type: str
        default: ''
    role:
        description:
          - The name or full Amazon Resource Name (ARN) of the IAM role that allows your Amazon ECS container agent to make calls to your load balancer
            on your behalf.
          - This parameter is only required if you are using a load balancer with your service in a network mode other than C(awsvpc).
        required: false
        type: str
        default: ''
    delay:
        description:
          - The time to wait before checking that the service is available.
        required: false
        default: 10
        type: int
    repeat:
        description:
          - The number of times to check that the service is available.
        required: false
        default: 10
        type: int
    force_new_deployment:
        description:
          - Force deployment of service even if there are no changes.
        required: false
        type: bool
        default: false
    deployment_controller:
        description:
          - The deployment controller to use for the service. If no deploymenet controller is specified, the ECS controller is used.
        required: false
        version_added: 4.1.0
        type: dict
        default: {}
        suboptions:
          type:
            type: str
            choices: ["ECS", "CODE_DEPLOY", "EXTERNAL"]
            description: The deployment controller type to use.
    deployment_configuration:
        description:
          - Optional parameters that control the deployment_configuration.
          - Format is '{"maximum_percent":<integer>, "minimum_healthy_percent":<integer>}
        required: false
        type: dict
        default: {}
        suboptions:
          maximum_percent:
            type: int
            description: Upper limit on the number of tasks in a service that are allowed in the RUNNING or PENDING state during a deployment.
          minimum_healthy_percent:
            type: int
            description: A lower limit on the number of tasks in a service that must remain in the RUNNING state during a deployment.
          deployment_circuit_breaker:
            type: dict
            description: The deployment circuit breaker determines whether a service deployment will fail if the service can't reach a steady state.
            suboptions:
              enable:
                type: bool
                description: If enabled, a service deployment will transition to a failed state and stop launching new tasks.
              rollback:
                type: bool
                description: If enabled, ECS will roll back your service to the last completed deployment after a failure.
    enable_execute_command:
        description:
          - Whether or not to enable the execute command functionality for the containers in the ECS task.
          - If I(enable_execute_command=true) execute command functionality is enabled on all containers in the ECS task.
        required: false
        type: bool
        version_added: 5.4.0
    placement_constraints:
        description:
          - The placement constraints for the tasks in the service.
          - See U(https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_PlacementConstraint.html) for more details.
        required: false
        type: list
        elements: dict
        default: []
        suboptions:
          type:
            description: The type of constraint.
            type: str
          expression:
            description: A cluster query language expression to apply to the constraint.
            required: false
            type: str
    purge_placement_constraints:
        version_added: 5.3.0
        description:
            - Toggle overwriting of existing placement constraints. This is needed for backwards compatibility.
            - By default I(purge_placement_constraints=false). In release 9.0.0 this will be changed to I(purge_placement_constraints=true).
        required: false
        type: bool
        default: false
    placement_strategy:
        description:
          - The placement strategy objects to use for tasks in your service. You can specify a maximum of 5 strategy rules per service.
        required: false
        type: list
        elements: dict
        default: []
        suboptions:
          type:
            description: The type of placement strategy.
            type: str
          field:
            description: The field to apply the placement strategy against.
            type: str
    purge_placement_strategy:
        version_added: 5.3.0
        description:
            - Toggle overwriting of existing placement strategy. This is needed for backwards compatibility.
            - By default I(purge_placement_strategy=false). In release 9.0.0 this will be changed to I(purge_placement_strategy=true).
        required: false
        type: bool
        default: false
    force_deletion:
        description:
          - Forcibly delete the service. Required when deleting a service with >0 scale, or no target group.
        default: False
        type: bool
        version_added: 2.1.0
    network_configuration:
        description:
          - Network configuration of the service. Only applicable for task definitions created with I(network_mode=awsvpc).
        type: dict
        suboptions:
          subnets:
            description:
              - A list of subnet IDs to associate with the task.
            type: list
            elements: str
          security_groups:
            description:
              - A list of security group names or group IDs to associate with the task.
            type: list
            elements: str
          assign_public_ip:
            description:
              - Whether the task's elastic network interface receives a public IP address.
            type: bool
    launch_type:
        description:
          - The launch type on which to run your service.
        required: false
        choices: ["EC2", "FARGATE"]
        type: str
    capacity_provider_strategy:
        version_added: 4.0.0
        description:
          - The capacity provider strategy to use with your service. You can specify a maximum of 6 providers per strategy.
        required: false
        type: list
        elements: dict
        default: []
        suboptions:
            capacity_provider:
                description:
                  - Name of capacity provider.
                type: str
            weight:
                description:
                  - The relative percentage of the total number of launched tasks that should use the specified provider.
                type: int
            base:
                description:
                  - How many tasks, at a minimum, should use the specified provider.
                type: int
    platform_version:
        type: str
        description:
          - Numeric part of platform version or LATEST
          - See U(https://docs.aws.amazon.com/AmazonECS/latest/developerguide/platform_versions.html) for more details.
        required: false
        version_added: 1.5.0
    health_check_grace_period_seconds:
        description:
          - Seconds to wait before health checking the freshly added/updated services.
        required: false
        type: int
    service_registries:
        description:
          - Describes service discovery registries this service will register with.
        type: list
        elements: dict
        default: []
        required: false
        suboptions:
            container_name:
                description:
                  - Container name for service discovery registration.
                type: str
            container_port:
                description:
                  - Container port for service discovery registration.
                type: int
            arn:
                description:
                  - Service discovery registry ARN.
                type: str
    scheduling_strategy:
        description:
          - The scheduling strategy.
          - Defaults to C(REPLICA) if not given to preserve previous behavior.
        required: false
        choices: ["DAEMON", "REPLICA"]
        type: str
    wait:
        description:
          - Whether or not to wait for the service to be inactive.
          - Waits only when I(state) is C(absent).
        type: bool
        default: false
        version_added: 4.1.0
    propagate_tags:
        description:
          - Propagate tags from ECS task defintition or ECS service to ECS task.
        required: false
        choices: ["TASK_DEFINITION", "SERVICE"]
        type: str
        version_added: 4.1.0
    tags:
        description:
          - A dictionary of tags to add or remove from the resource.
        type: dict
        required: false
        version_added: 4.1.0
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.
# Basic provisioning example
- community.aws.ecs_service:
    state: present
    name: console-test-service
    cluster: new_cluster
    task_definition: 'new_cluster-task:1'
    desired_count: 0

- name: create ECS service on VPC network
  community.aws.ecs_service:
    state: present
    name: console-test-service
    cluster: new_cluster
    task_definition: 'new_cluster-task:1'
    desired_count: 0
    network_configuration:
      subnets:
        - subnet-abcd1234
      security_groups:
        - sg-aaaa1111
        - my_security_group

# Simple example to delete
- community.aws.ecs_service:
    name: default
    state: absent
    cluster: new_cluster

# With custom deployment configuration (added in version 2.3), placement constraints and strategy (added in version 2.4)
- community.aws.ecs_service:
    state: present
    name: test-service
    cluster: test-cluster
    task_definition: test-task-definition
    desired_count: 3
    deployment_configuration:
      minimum_healthy_percent: 75
      maximum_percent: 150
    placement_constraints:
      - type: memberOf
        expression: 'attribute:flavor==test'
    placement_strategy:
      - type: binpack
        field: memory

# With deployment circuit breaker (added in version 4.0)
- community.aws.ecs_service:
    state: present
    name: test-service
    cluster: test-cluster
    task_definition: test-task-definition
    desired_count: 3
    deployment_configuration:
      deployment_circuit_breaker:
        enable: true
        rollback: true

# With capacity_provider_strategy (added in version 4.0)
- community.aws.ecs_service:
    state: present
    name: test-service
    cluster: test-cluster
    task_definition: test-task-definition
    desired_count: 1
    capacity_provider_strategy:
      - capacity_provider: test-capacity-provider-1
        weight: 1
        base: 0

# With tags and tag propagation
- community.aws.ecs_service:
    state: present
    name: tags-test-service
    cluster: new_cluster
    task_definition: 'new_cluster-task:1'
    desired_count: 1
    tags:
      Firstname: jane
      lastName: doe
    propagate_tags: SERVICE
"""

RETURN = r"""
service:
    description: Details of created service.
    returned: when creating a service
    type: complex
    contains:
        capacityProviderStrategy:
            version_added: 4.0.0
            description: The capacity provider strategy to use with your service.
            returned: always
            type: complex
            contains:
                base:
                    description: How many tasks, at a minimum, should use the specified provider.
                    returned: always
                    type: int
                capacityProvider:
                    description: Name of capacity provider.
                    returned: always
                    type: str
                weight:
                    description: The relative percentage of the total number of launched tasks that should use the specified provider.
                    returned: always
                    type: int
        clusterArn:
            description: The Amazon Resource Name (ARN) of the of the cluster that hosts the service.
            returned: always
            type: str
        desiredCount:
            description: The desired number of instantiations of the task definition to keep running on the service.
            returned: always
            type: int
        loadBalancers:
            description:
                - A list of load balancer objects
            returned: always
            type: complex
            contains:
                loadBalancerName:
                    description: the name
                    returned: always
                    type: str
                containerName:
                    description: The name of the container to associate with the load balancer.
                    returned: always
                    type: str
                containerPort:
                    description: The port on the container to associate with the load balancer.
                    returned: always
                    type: int
        pendingCount:
            description: The number of tasks in the cluster that are in the PENDING state.
            returned: always
            type: int
        runningCount:
            description: The number of tasks in the cluster that are in the RUNNING state.
            returned: always
            type: int
        serviceArn:
            description:
                - The Amazon Resource Name (ARN) that identifies the service. The ARN contains the C(arn:aws:ecs) namespace, followed by
                  the region of the service, the AWS account ID of the service owner, the service namespace, and then the service name.
            sample: 'arn:aws:ecs:us-east-1:123456789012:service/my-service'
            returned: always
            type: str
        serviceName:
            description: A user-generated string used to identify the service
            returned: always
            type: str
        status:
            description: The valid values are ACTIVE, DRAINING, or INACTIVE.
            returned: always
            type: str
        tags:
            description: The tags applied to this resource.
            returned: success
            type: dict
        taskDefinition:
            description: The ARN of a task definition to use for tasks in the service.
            returned: always
            type: str
        deployments:
            description: list of service deployments
            returned: always
            type: list
            elements: dict
        deploymentConfiguration:
            description: dictionary of deploymentConfiguration
            returned: always
            type: complex
            contains:
                maximumPercent:
                    description: maximumPercent param
                    returned: always
                    type: int
                minimumHealthyPercent:
                    description: minimumHealthyPercent param
                    returned: always
                    type: int
                deploymentCircuitBreaker:
                    description: dictionary of deploymentCircuitBreaker
                    returned: always
                    type: complex
                    contains:
                        enable:
                            description: The state of the circuit breaker feature.
                            returned: always
                            type: bool
                        rollback:
                            description: The state of the rollback feature of the circuit breaker.
                            returned: always
                            type: bool
        events:
            description: list of service events
            returned: always
            type: list
            elements: dict
        placementConstraints:
            description: List of placement constraints objects
            returned: always
            type: list
            elements: dict
            contains:
                type:
                    description: The type of constraint. Valid values are distinctInstance and memberOf.
                    returned: always
                    type: str
                expression:
                    description: A cluster query language expression to apply to the constraint. Note you cannot specify an expression if the constraint type is
                                 distinctInstance.
                    returned: always
                    type: str
        placementStrategy:
            description: List of placement strategy objects
            returned: always
            type: list
            elements: dict
            contains:
                type:
                    description: The type of placement strategy. Valid values are random, spread and binpack.
                    returned: always
                    type: str
                field:
                    description: The field to apply the placement strategy against. For the spread placement strategy, valid values are instanceId
                                 (or host, which has the same effect), or any platform or custom attribute that is applied to a container instance,
                                 such as attribute:ecs.availability-zone. For the binpack placement strategy, valid values are CPU and MEMORY.
                    returned: always
                    type: str
        propagateTags:
            description: The type of tag propagation applied to the resource.
            returned: always
            type: str
ansible_facts:
    description: Facts about deleted service.
    returned: when deleting a service
    type: complex
    contains:
        service:
            description: Details of deleted service.
            returned: when service existed and was deleted
            type: complex
            contains:
                clusterArn:
                    description: The Amazon Resource Name (ARN) of the of the cluster that hosts the service.
                    returned: always
                    type: str
                desiredCount:
                    description: The desired number of instantiations of the task definition to keep running on the service.
                    returned: always
                    type: int
                loadBalancers:
                    description: A list of load balancer objects
                    returned: always
                    type: complex
                    contains:
                        loadBalancerName:
                            description: the name
                            returned: always
                            type: str
                        containerName:
                            description: The name of the container to associate with the load balancer.
                            returned: always
                            type: str
                        containerPort:
                            description: The port on the container to associate with the load balancer.
                            returned: always
                            type: int
                pendingCount:
                    description: The number of tasks in the cluster that are in the PENDING state.
                    returned: always
                    type: int
                runningCount:
                    description: The number of tasks in the cluster that are in the RUNNING state.
                    returned: always
                    type: int
                serviceArn:
                    description:
                        - The Amazon Resource Name (ARN) that identifies the service. The ARN contains the arn:aws:ecs namespace, followed by the region
                          of the service, the AWS account ID of the service owner, the service namespace, and then the service name.
                    sample: 'arn:aws:ecs:us-east-1:123456789012:service/my-service'
                    returned: always
                    type: str
                serviceName:
                    description: A user-generated string used to identify the service
                    returned: always
                    type: str
                status:
                    description: The valid values are ACTIVE, DRAINING, or INACTIVE.
                    returned: always
                    type: str
                tags:
                    description: The tags applied to this resource.
                    returned: when tags found
                    type: list
                    elements: dict
                taskDefinition:
                    description: The ARN of a task definition to use for tasks in the service.
                    returned: always
                    type: str
                deployments:
                    description: list of service deployments
                    returned: always
                    type: list
                    elements: dict
                deploymentConfiguration:
                    description: dictionary of deploymentConfiguration
                    returned: always
                    type: complex
                    contains:
                        maximumPercent:
                            description: maximumPercent param
                            returned: always
                            type: int
                        minimumHealthyPercent:
                            description: minimumHealthyPercent param
                            returned: always
                            type: int
                        deploymentCircuitBreaker:
                            description: dictionary of deploymentCircuitBreaker
                            returned: always
                            type: complex
                            contains:
                                enable:
                                    description: The state of the circuit breaker feature.
                                    returned: always
                                    type: bool
                                rollback:
                                    description: The state of the rollback feature of the circuit breaker.
                                    returned: always
                                    type: bool
                events:
                    description: list of service events
                    returned: always
                    type: list
                    elements: dict
                placementConstraints:
                    description: List of placement constraints objects
                    returned: always
                    type: list
                    elements: dict
                    contains:
                        type:
                            description: The type of constraint. Valid values are distinctInstance and memberOf.
                            returned: always
                            type: str
                        expression:
                            description: A cluster query language expression to apply to the constraint. Note you cannot specify an expression if
                                         the constraint type is distinctInstance.
                            returned: always
                            type: str
                placementStrategy:
                    description: List of placement strategy objects
                    returned: always
                    type: list
                    elements: dict
                    contains:
                        type:
                            description: The type of placement strategy. Valid values are random, spread and binpack.
                            returned: always
                            type: str
                        field:
                            description: The field to apply the placement strategy against. For the spread placement strategy, valid values are instanceId
                                         (or host, which has the same effect), or any platform or custom attribute that is applied to a container instance,
                                         such as attribute:ecs.availability-zone. For the binpack placement strategy, valid values are CPU and MEMORY.
                            returned: always
                            type: str
                propagateTags:
                    description: The type of tag propagation applied to the resource
                    returned: always
                    type: str

"""

import time

try:
    import botocore
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import snake_dict_to_camel_dict

from ansible_collections.amazon.aws.plugins.module_utils.ec2 import get_ec2_security_group_ids_from_names
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.transformation import map_complex_type

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule

DEPLOYMENT_CONTROLLER_TYPE_MAP = {
    "type": "str",
}

DEPLOYMENT_CONFIGURATION_TYPE_MAP = {
    "maximum_percent": "int",
    "minimum_healthy_percent": "int",
    "deployment_circuit_breaker": "dict",
}


class EcsServiceManager:
    """Handles ECS Services"""

    def __init__(self, module):
        self.module = module
        self.ecs = module.client("ecs")
        self.ec2 = module.client("ec2")

    def format_network_configuration(self, network_config):
        result = dict()
        if network_config["subnets"] is not None:
            result["subnets"] = network_config["subnets"]
        else:
            self.module.fail_json(msg="Network configuration must include subnets")
        if network_config["security_groups"] is not None:
            groups = network_config["security_groups"]
            if any(not sg.startswith("sg-") for sg in groups):
                try:
                    vpc_id = self.ec2.describe_subnets(SubnetIds=[result["subnets"][0]])["Subnets"][0]["VpcId"]
                    groups = get_ec2_security_group_ids_from_names(groups, self.ec2, vpc_id)
                except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
                    self.module.fail_json_aws(e, msg="Couldn't look up security groups")
            result["securityGroups"] = groups
        if network_config["assign_public_ip"] is not None:
            if network_config["assign_public_ip"] is True:
                result["assignPublicIp"] = "ENABLED"
            else:
                result["assignPublicIp"] = "DISABLED"
        return dict(awsvpcConfiguration=result)

    def find_in_array(self, array_of_services, service_name, field_name="serviceArn"):
        for c in array_of_services:
            if c[field_name].endswith(service_name):
                return c
        return None

    def describe_service(self, cluster_name, service_name):
        response = self.ecs.describe_services(
            cluster=cluster_name,
            services=[service_name],
            include=["TAGS"],
        )
        msg = ""

        if len(response["failures"]) > 0:
            c = self.find_in_array(response["failures"], service_name, "arn")
            msg += ", failure reason is " + c["reason"]
            if c and c["reason"] == "MISSING":
                return None
            # fall thru and look through found ones
        if len(response["services"]) > 0:
            c = self.find_in_array(response["services"], service_name)
            if c:
                return c
        raise Exception(f"Unknown problem describing service {service_name}.")

    def is_matching_service(self, expected, existing):
        # aws returns the arn of the task definition
        #   arn:aws:ecs:eu-central-1:123456789:task-definition/ansible-fargate-nginx:3
        # but the user is just entering
        #   ansible-fargate-nginx:3
        if expected["task_definition"] != existing["taskDefinition"].split("/")[-1]:
            if existing.get("deploymentController", {}).get("type", None) != "CODE_DEPLOY":
                return False

        if expected.get("health_check_grace_period_seconds"):
            if expected.get("health_check_grace_period_seconds") != existing.get("healthCheckGracePeriodSeconds"):
                return False

        if (expected["load_balancers"] or []) != existing["loadBalancers"]:
            return False

        if (expected["propagate_tags"] or "NONE") != existing["propagateTags"]:
            return False

        if boto3_tag_list_to_ansible_dict(existing.get("tags", [])) != (expected["tags"] or {}):
            return False

        if (expected["enable_execute_command"] or False) != existing.get("enableExecuteCommand", False):
            return False

        # expected is params. DAEMON scheduling strategy returns desired count equal to
        # number of instances running; don't check desired count if scheduling strat is daemon
        if expected["scheduling_strategy"] != "DAEMON":
            if (expected["desired_count"] or 0) != existing["desiredCount"]:
                return False

        return True

    def create_service(
        self,
        service_name,
        cluster_name,
        task_definition,
        load_balancers,
        desired_count,
        client_token,
        role,
        deployment_controller,
        deployment_configuration,
        placement_constraints,
        placement_strategy,
        health_check_grace_period_seconds,
        network_configuration,
        service_registries,
        launch_type,
        platform_version,
        scheduling_strategy,
        capacity_provider_strategy,
        tags,
        propagate_tags,
        enable_execute_command,
    ):
        params = dict(
            cluster=cluster_name,
            serviceName=service_name,
            taskDefinition=task_definition,
            loadBalancers=load_balancers,
            clientToken=client_token,
            role=role,
            deploymentConfiguration=deployment_configuration,
            placementStrategy=placement_strategy,
        )
        if network_configuration:
            params["networkConfiguration"] = network_configuration
        if deployment_controller:
            params["deploymentController"] = deployment_controller
        if launch_type:
            params["launchType"] = launch_type
        if platform_version:
            params["platformVersion"] = platform_version
        if self.health_check_setable(params) and health_check_grace_period_seconds is not None:
            params["healthCheckGracePeriodSeconds"] = health_check_grace_period_seconds
        if service_registries:
            params["serviceRegistries"] = service_registries

        # filter placement_constraint and left only those where value is not None
        # use-case: `distinctInstance` type should never contain `expression`, but None will fail `str` type validation
        if placement_constraints:
            params["placementConstraints"] = [
                {key: value for key, value in constraint.items() if value is not None}
                for constraint in placement_constraints
            ]

        # desired count is not required if scheduling strategy is daemon
        if desired_count is not None:
            params["desiredCount"] = desired_count
        if capacity_provider_strategy:
            params["capacityProviderStrategy"] = capacity_provider_strategy
        if propagate_tags:
            params["propagateTags"] = propagate_tags
        # desired count is not required if scheduling strategy is daemon
        if desired_count is not None:
            params["desiredCount"] = desired_count
        if tags:
            params["tags"] = ansible_dict_to_boto3_tag_list(tags, "key", "value")

        if scheduling_strategy:
            params["schedulingStrategy"] = scheduling_strategy
        if enable_execute_command:
            params["enableExecuteCommand"] = enable_execute_command

        response = self.ecs.create_service(**params)
        return self.jsonize(response["service"])

    def update_service(
        self,
        service_name,
        cluster_name,
        task_definition,
        desired_count,
        deployment_configuration,
        placement_constraints,
        placement_strategy,
        network_configuration,
        health_check_grace_period_seconds,
        force_new_deployment,
        capacity_provider_strategy,
        load_balancers,
        purge_placement_constraints,
        purge_placement_strategy,
        enable_execute_command,
    ):
        params = dict(
            cluster=cluster_name,
            service=service_name,
            taskDefinition=task_definition,
            deploymentConfiguration=deployment_configuration,
        )
        # filter placement_constraint and left only those where value is not None
        # use-case: `distinctInstance` type should never contain `expression`, but None will fail `str` type validation
        if placement_constraints:
            params["placementConstraints"] = [
                {key: value for key, value in constraint.items() if value is not None}
                for constraint in placement_constraints
            ]

        if purge_placement_constraints and not placement_constraints:
            params["placementConstraints"] = []

        if placement_strategy:
            params["placementStrategy"] = placement_strategy

        if purge_placement_strategy and not placement_strategy:
            params["placementStrategy"] = []

        if network_configuration:
            params["networkConfiguration"] = network_configuration
        if force_new_deployment:
            params["forceNewDeployment"] = force_new_deployment
        if capacity_provider_strategy:
            params["capacityProviderStrategy"] = capacity_provider_strategy
        if health_check_grace_period_seconds is not None:
            params["healthCheckGracePeriodSeconds"] = health_check_grace_period_seconds
        # desired count is not required if scheduling strategy is daemon
        if desired_count is not None:
            params["desiredCount"] = desired_count
        if enable_execute_command is not None:
            params["enableExecuteCommand"] = enable_execute_command

        if load_balancers:
            params["loadBalancers"] = load_balancers

        response = self.ecs.update_service(**params)

        return self.jsonize(response["service"])

    def jsonize(self, service):
        # some fields are datetime which is not JSON serializable
        # make them strings
        if "createdAt" in service:
            service["createdAt"] = str(service["createdAt"])
        if "deployments" in service:
            for d in service["deployments"]:
                if "createdAt" in d:
                    d["createdAt"] = str(d["createdAt"])
                if "updatedAt" in d:
                    d["updatedAt"] = str(d["updatedAt"])
        if "events" in service:
            for e in service["events"]:
                if "createdAt" in e:
                    e["createdAt"] = str(e["createdAt"])
        return service

    def delete_service(self, service, cluster=None, force=False):
        return self.ecs.delete_service(cluster=cluster, service=service, force=force)

    def health_check_setable(self, params):
        load_balancers = params.get("loadBalancers", [])
        return len(load_balancers) > 0


def main():
    argument_spec = dict(
        state=dict(required=True, choices=["present", "absent", "deleting"]),
        name=dict(required=True, type="str", aliases=["service"]),
        cluster=dict(required=False, type="str", default="default"),
        task_definition=dict(required=False, type="str"),
        load_balancers=dict(required=False, default=[], type="list", elements="dict"),
        desired_count=dict(required=False, type="int"),
        client_token=dict(required=False, default="", type="str", no_log=False),
        role=dict(required=False, default="", type="str"),
        delay=dict(required=False, type="int", default=10),
        repeat=dict(required=False, type="int", default=10),
        force_new_deployment=dict(required=False, default=False, type="bool"),
        force_deletion=dict(required=False, default=False, type="bool"),
        deployment_controller=dict(required=False, default={}, type="dict"),
        deployment_configuration=dict(required=False, default={}, type="dict"),
        wait=dict(required=False, default=False, type="bool"),
        placement_constraints=dict(
            required=False,
            default=[],
            type="list",
            elements="dict",
            options=dict(type=dict(type="str"), expression=dict(required=False, type="str")),
        ),
        purge_placement_constraints=dict(required=False, default=False, type="bool"),
        placement_strategy=dict(
            required=False,
            default=[],
            type="list",
            elements="dict",
            options=dict(
                type=dict(type="str"),
                field=dict(type="str"),
            ),
        ),
        purge_placement_strategy=dict(required=False, default=False, type="bool"),
        health_check_grace_period_seconds=dict(required=False, type="int"),
        network_configuration=dict(
            required=False,
            type="dict",
            options=dict(
                subnets=dict(type="list", elements="str"),
                security_groups=dict(type="list", elements="str"),
                assign_public_ip=dict(type="bool"),
            ),
        ),
        launch_type=dict(required=False, choices=["EC2", "FARGATE"]),
        platform_version=dict(required=False, type="str"),
        service_registries=dict(required=False, type="list", default=[], elements="dict"),
        scheduling_strategy=dict(required=False, choices=["DAEMON", "REPLICA"]),
        capacity_provider_strategy=dict(
            required=False,
            type="list",
            default=[],
            elements="dict",
            options=dict(
                capacity_provider=dict(type="str"),
                weight=dict(type="int"),
                base=dict(type="int"),
            ),
        ),
        propagate_tags=dict(required=False, choices=["TASK_DEFINITION", "SERVICE"]),
        tags=dict(required=False, type="dict"),
        enable_execute_command=dict(required=False, type="bool"),
    )

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[("launch_type", "FARGATE", ["network_configuration"])],
        required_together=[["load_balancers", "role"]],
        mutually_exclusive=[["launch_type", "capacity_provider_strategy"]],
    )

    if module.params["state"] == "present":
        if module.params["scheduling_strategy"] == "REPLICA" and module.params["desired_count"] is None:
            module.fail_json(msg="state is present, scheduling_strategy is REPLICA; missing desired_count")
        if module.params["task_definition"] is None and not module.params["force_new_deployment"]:
            module.fail_json(msg="Either task_definition or force_new_deployment is required when status is present.")

    if len(module.params["capacity_provider_strategy"]) > 6:
        module.fail_json(msg="AWS allows a maximum of six capacity providers in the strategy.")

    service_mgr = EcsServiceManager(module)
    if module.params["network_configuration"]:
        network_configuration = service_mgr.format_network_configuration(module.params["network_configuration"])
    else:
        network_configuration = None

    deployment_controller = map_complex_type(module.params["deployment_controller"], DEPLOYMENT_CONTROLLER_TYPE_MAP)

    deploymentController = snake_dict_to_camel_dict(deployment_controller)

    deployment_configuration = map_complex_type(
        module.params["deployment_configuration"], DEPLOYMENT_CONFIGURATION_TYPE_MAP
    )

    deploymentConfiguration = snake_dict_to_camel_dict(deployment_configuration)
    serviceRegistries = list(map(snake_dict_to_camel_dict, module.params["service_registries"]))
    capacityProviders = list(map(snake_dict_to_camel_dict, module.params["capacity_provider_strategy"]))

    try:
        existing = service_mgr.describe_service(module.params["cluster"], module.params["name"])
    except Exception as e:
        module.fail_json_aws(
            e,
            msg=f"Exception describing service '{module.params['name']}' in cluster '{module.params['cluster']}'",
        )

    results = dict(changed=False)

    if module.params["state"] == "present":
        matching = False
        update = False

        if existing and "status" in existing and existing["status"] == "ACTIVE":
            if module.params["force_new_deployment"]:
                update = True
            elif service_mgr.is_matching_service(module.params, existing):
                matching = True
                results["service"] = existing
            else:
                update = True

        if not matching:
            if not module.check_mode:
                role = module.params["role"]
                clientToken = module.params["client_token"]

                loadBalancers = []
                for loadBalancer in module.params["load_balancers"]:
                    if "containerPort" in loadBalancer:
                        loadBalancer["containerPort"] = int(loadBalancer["containerPort"])
                    loadBalancers.append(loadBalancer)

                for loadBalancer in loadBalancers:
                    if "containerPort" in loadBalancer:
                        loadBalancer["containerPort"] = int(loadBalancer["containerPort"])

                if update:
                    # check various parameters and AWS SDK versions and give a helpful error if the SDK is not new enough for feature

                    if module.params["scheduling_strategy"]:
                        if (existing["schedulingStrategy"]) != module.params["scheduling_strategy"]:
                            module.fail_json(
                                msg="It is not possible to update the scheduling strategy of an existing service"
                            )

                    if module.params["service_registries"]:
                        if (existing["serviceRegistries"] or []) != serviceRegistries:
                            module.fail_json(
                                msg="It is not possible to update the service registries of an existing service"
                            )
                    if module.params["capacity_provider_strategy"]:
                        if "launchType" in existing.keys():
                            module.fail_json(
                                msg="It is not possible to change an existing service from launch_type to capacity_provider_strategy."
                            )
                    if module.params["launch_type"]:
                        if "capacityProviderStrategy" in existing.keys():
                            module.fail_json(
                                msg="It is not possible to change an existing service from capacity_provider_strategy to launch_type."
                            )
                    if (existing["loadBalancers"] or []) != loadBalancers:
                        # fails if deployment type is not CODE_DEPLOY or ECS
                        if existing["deploymentController"]["type"] not in ["CODE_DEPLOY", "ECS"]:
                            module.fail_json(
                                msg="It is not possible to update the load balancers of an existing service"
                            )

                    if existing.get("deploymentController", {}).get("type", None) == "CODE_DEPLOY":
                        task_definition = ""
                        network_configuration = []
                    else:
                        task_definition = module.params["task_definition"]

                    if module.params["propagate_tags"] and module.params["propagate_tags"] != existing["propagateTags"]:
                        module.fail_json(
                            msg="It is not currently supported to enable propagation tags of an existing service"
                        )

                    if (
                        module.params["tags"]
                        and boto3_tag_list_to_ansible_dict(existing["tags"]) != module.params["tags"]
                    ):
                        module.fail_json(msg="It is not currently supported to change tags of an existing service")

                    updatedLoadBalancers = loadBalancers if existing["deploymentController"]["type"] == "ECS" else []

                    if task_definition is None and module.params["force_new_deployment"]:
                        task_definition = existing["taskDefinition"]

                    try:
                        # update required
                        response = service_mgr.update_service(
                            module.params["name"],
                            module.params["cluster"],
                            task_definition,
                            module.params["desired_count"],
                            deploymentConfiguration,
                            module.params["placement_constraints"],
                            module.params["placement_strategy"],
                            network_configuration,
                            module.params["health_check_grace_period_seconds"],
                            module.params["force_new_deployment"],
                            capacityProviders,
                            updatedLoadBalancers,
                            module.params["purge_placement_constraints"],
                            module.params["purge_placement_strategy"],
                            module.params["enable_execute_command"],
                        )
                    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
                        module.fail_json_aws(e, msg="Couldn't create service")

                else:
                    try:
                        response = service_mgr.create_service(
                            module.params["name"],
                            module.params["cluster"],
                            module.params["task_definition"],
                            loadBalancers,
                            module.params["desired_count"],
                            clientToken,
                            role,
                            deploymentController,
                            deploymentConfiguration,
                            module.params["placement_constraints"],
                            module.params["placement_strategy"],
                            module.params["health_check_grace_period_seconds"],
                            network_configuration,
                            serviceRegistries,
                            module.params["launch_type"],
                            module.params["platform_version"],
                            module.params["scheduling_strategy"],
                            capacityProviders,
                            module.params["tags"],
                            module.params["propagate_tags"],
                            module.params["enable_execute_command"],
                        )
                    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
                        module.fail_json_aws(e, msg="Couldn't create service")

                if response.get("tags", None):
                    response["tags"] = boto3_tag_list_to_ansible_dict(response["tags"])
                results["service"] = response

            results["changed"] = True

    elif module.params["state"] == "absent":
        if not existing:
            pass
        else:
            # it exists, so we should delete it and mark changed.
            # return info about the cluster deleted
            del existing["deployments"]
            del existing["events"]
            results["ansible_facts"] = existing
            if "status" in existing and existing["status"] == "INACTIVE":
                results["changed"] = False
            else:
                if not module.check_mode:
                    try:
                        service_mgr.delete_service(
                            module.params["name"],
                            module.params["cluster"],
                            module.params["force_deletion"],
                        )

                        # Wait for service to be INACTIVE prior to exiting
                        if module.params["wait"]:
                            waiter = service_mgr.ecs.get_waiter("services_inactive")
                            try:
                                waiter.wait(
                                    services=[module.params["name"]],
                                    cluster=module.params["cluster"],
                                    WaiterConfig={
                                        "Delay": module.params["delay"],
                                        "MaxAttempts": module.params["repeat"],
                                    },
                                )
                            except botocore.exceptions.WaiterError as e:
                                module.fail_json_aws(e, "Timeout waiting for service removal")
                    except botocore.exceptions.ClientError as e:
                        module.fail_json_aws(e, msg="Couldn't delete service")

                results["changed"] = True

    elif module.params["state"] == "deleting":
        if not existing:
            module.fail_json(msg="Service '" + module.params["name"] + " not found.")

        # it exists, so we should delete it and mark changed.
        # return info about the cluster deleted
        delay = module.params["delay"]
        repeat = module.params["repeat"]
        time.sleep(delay)
        for i in range(repeat):
            existing = service_mgr.describe_service(module.params["cluster"], module.params["name"])
            status = existing["status"]
            if status == "INACTIVE":
                results["changed"] = True
                break
            time.sleep(delay)
        if i is repeat - 1:
            module.fail_json(msg=f"Service still not deleted after {repeat} tries of {delay} seconds each.")

    module.exit_json(**results)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2018 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: acme_account_info
author: "Felix Fontein (@felixfontein)"
short_description: Retrieves information on ACME accounts
description:
  - Allows to retrieve information on accounts a CA supporting the L(ACME protocol,https://tools.ietf.org/html/rfc8555), such
    as L(Let's Encrypt,https://letsencrypt.org/).
  - This module only works with the ACME v2 protocol.
notes:
  - The M(community.crypto.acme_account) module allows to modify, create and delete ACME accounts.
  - This module was called C(acme_account_facts) before Ansible 2.8. The usage did not change.
extends_documentation_fragment:
  - community.crypto._acme.basic
  - community.crypto._acme.account
  - community.crypto._attributes
  - community.crypto._attributes.actiongroup_acme
  - community.crypto._attributes.info_module
  - community.crypto._attributes.idempotent_not_modify_state
options:
  retrieve_orders:
    description:
      - Whether to retrieve the list of order URLs or order objects, if provided by the ACME server.
      - A value of V(ignore) will not fetch the list of orders.
      - If the value is not V(ignore) and the ACME server supports orders, the RV(order_uris) return value is always populated.
        The RV(orders) return value is only returned if this option is set to V(object_list).
      - Currently, Let's Encrypt does not return orders, so the RV(orders) result will always be empty.
    type: str
    choices:
      - ignore
      - url_list
      - object_list
    default: ignore
seealso:
  - module: community.crypto.acme_account
    description: Allows to create, modify or delete an ACME account.
"""

EXAMPLES = r"""
---
- name: Check whether an account with the given account key exists
  community.crypto.acme_account_info:
    account_key_src: /etc/pki/cert/private/account.key
  register: account_data
- name: Verify that account exists
  ansible.builtin.assert:
    that:
      - account_data.exists
- name: Print account URI
  ansible.builtin.debug:
    var: account_data.account_uri
- name: Print account contacts
  ansible.builtin.debug:
    var: account_data.account.contact

- name: Check whether the account exists and is accessible with the given account key
  acme_account_info:
    account_key_content: "{{ acme_account_key }}"
    account_uri: "{{ acme_account_uri }}"
  register: account_data
- name: Verify that account exists
  ansible.builtin.assert:
    that:
      - account_data.exists
- name: Print account contacts
  ansible.builtin.debug:
    var: account_data.account.contact
"""

RETURN = r"""
exists:
  description: Whether the account exists.
  returned: always
  type: bool

account_uri:
  description: ACME account URI, or None if account does not exist.
  returned: always
  type: str

account:
  description: The account information, as retrieved from the ACME server.
  returned: if account exists
  type: dict
  contains:
    contact:
      description: The challenge resource that must be created for validation.
      returned: always
      type: list
      elements: str
      sample: ['mailto:me@example.com', 'tel:00123456789']
    status:
      description: The account's status.
      returned: always
      type: str
      choices: ['valid', 'deactivated', 'revoked']
      sample: valid
    orders:
      description:
        - A URL where a list of orders can be retrieved for this account.
        - Use the O(retrieve_orders) option to query this URL and retrieve the complete list of orders.
      returned: always
      type: str
      sample: https://example.ca/account/1/orders
    public_account_key:
      description: The public account key as a L(JSON Web Key,https://tools.ietf.org/html/rfc7517).
      returned: always
      type: str
      sample: '{"kty":"EC","crv":"P-256","x":"MKBCTNIcKUSDii11ySs3526iDZ8AiTo7Tu6KPAqv7D4","y":"4Etl6SRW2YiLUrN5vfvVHuhp7x8PxltmWWlbbM4IFyM"}'

orders:
  description:
    - The list of orders.
  type: list
  elements: dict
  returned: if account exists, O(retrieve_orders) is V(object_list), and server supports order listing
  contains:
    status:
      description: The order's status.
      type: str
      choices:
        - pending
        - ready
        - processing
        - valid
        - invalid
    expires:
      description:
        - When the order expires.
        - Timestamp should be formatted as described in RFC3339.
        - Only required to be included in result when RV(orders[].status) is V(pending) or V(valid).
      type: str
      returned: when server gives expiry date
    identifiers:
      description:
        - List of identifiers this order is for.
      type: list
      elements: dict
      contains:
        type:
          description: Type of identifier.
          type: str
          choices:
            - dns
            - ip
        value:
          description: Name of identifier. Hostname or IP address.
          type: str
        wildcard:
          description: "Whether RV(orders[].identifiers[].value) is actually a wildcard. The wildcard prefix C(*.) is not
            included in RV(orders[].identifiers[].value) if this is V(true)."
          type: bool
          returned: required to be included if the identifier is wildcarded
    notBefore:
      description:
        - The requested value of the C(notBefore) field in the certificate.
        - Date should be formatted as described in RFC3339.
        - Server is not required to return this.
      type: str
      returned: when server returns this
    notAfter:
      description:
        - The requested value of the C(notAfter) field in the certificate.
        - Date should be formatted as described in RFC3339.
        - Server is not required to return this.
      type: str
      returned: when server returns this
    error:
      description:
        - In case an error occurred during processing, this contains information about the error.
        - The field is structured as a problem document (RFC7807).
      type: dict
      returned: when an error occurred
    authorizations:
      description:
        - A list of URLs for authorizations for this order.
      type: list
      elements: str
    finalize:
      description:
        - A URL used for finalizing an ACME order.
      type: str
    certificate:
      description:
        - The URL for retrieving the certificate.
      type: str
      returned: when certificate was issued

order_uris:
  description:
    - The list of orders.
    - If O(retrieve_orders) is V(url_list), this will be a list of URLs.
    - If O(retrieve_orders) is V(object_list), this will be a list of objects.
  type: list
  elements: str
  returned: if account exists, O(retrieve_orders) is not V(ignore), and server supports order listing
  version_added: 1.5.0
"""

import typing as t

from ansible_collections.community.crypto.plugins.module_utils._acme.account import (
    ACMEAccount,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (
    ACMEClient,
    create_backend,
    create_default_argspec,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ACMEProtocolException,
    ModuleFailException,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.utils import (
    process_links,
)

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover


def _collect_next(info: dict[str, t.Any]) -> list[str]:
    result: list[str] = []

    def f(link: str, relation: str) -> None:
        if relation == "next":
            result.append(link)

    process_links(info=info, callback=f)
    return result


def get_orders_list(
    module: AnsibleModule, client: ACMEClient, orders_url: str
) -> list[str]:
    """
    Retrieves order URL list (handles pagination).
    """
    orders: list[str] = []
    next_orders_url: str | None = orders_url
    while next_orders_url:
        # Get part of orders list
        res, info = client.get_request(
            next_orders_url, parse_json_result=True, fail_on_error=True
        )
        if not isinstance(res, dict):
            raise ACMEProtocolException(
                module=module,
                msg="Unexpected account information",
                info=info,
                content_json=res,
            )
        if not res.get("orders"):
            if orders:
                module.warn(
                    f"When retrieving orders list part {next_orders_url}, got empty result list"
                )
            break
        # Add order URLs to result list
        orders.extend(res["orders"])
        # Extract URL of next part of results list
        new_orders_url: list[str | None] = []
        new_orders_url.extend(_collect_next(info))
        new_orders_url.append(None)
        previous_orders_url, next_orders_url = next_orders_url, new_orders_url.pop(0)
        if next_orders_url == previous_orders_url:
            # Prevent infinite loop
            next_orders_url = None
    return orders


def get_order(client: ACMEClient, order_url: str) -> dict[str, t.Any]:
    """
    Retrieve order data.
    """
    result, info = client.get_request(
        order_url, parse_json_result=True, fail_on_error=True
    )
    if not isinstance(result, dict):
        raise ACMEProtocolException(
            module=client.module,
            msg="Unexpected order data",
            info=info,
            content_json=result,
        )
    return result


def main() -> t.NoReturn:
    argument_spec = create_default_argspec()
    argument_spec.update_argspec(
        retrieve_orders={
            "type": "str",
            "default": "ignore",
            "choices": ["ignore", "url_list", "object_list"],
        },
    )
    module = argument_spec.create_ansible_module(supports_check_mode=True)
    backend = create_backend(module, needs_acme_v2=True)

    try:
        client = ACMEClient(module=module, backend=backend)
        account = ACMEAccount(client=client)
        # Check whether account exists
        created, account_data = account.setup_account(
            contact=[],
            allow_creation=False,
            remove_account_uri_if_not_exists=True,
        )
        if created:
            raise AssertionError("Unwanted account creation")  # pragma: no cover
        result: dict[str, t.Any] = {
            "changed": False,
            "exists": False,
            "account_uri": None,
        }
        if client.account_uri is not None and account_data:
            result["account_uri"] = client.account_uri
            result["exists"] = True
            # Make sure promised data is there
            account_data_dict = dict(account_data)
            if "contact" not in account_data:
                account_data_dict["contact"] = []
            if client.account_key_data:
                account_data_dict["public_account_key"] = client.account_key_data["jwk"]
            result["account"] = account_data_dict
            # Retrieve orders list
            if (
                account_data.get("orders")
                and module.params["retrieve_orders"] != "ignore"
            ):
                orders = get_orders_list(module, client, account_data["orders"])
                result["order_uris"] = orders
                if module.params["retrieve_orders"] == "object_list":
                    result["orders"] = [get_order(client, order) for order in orders]
        module.exit_json(**result)
    except ModuleFailException as e:
        e.do_fail(module=module)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2019 Zim Kalinowski, (@zikalino)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_gallery
version_added: "0.1.2"
short_description: Manage Azure Shared Image Gallery instance
description:
    - Create, update and delete instance of Azure Shared Image Gallery (SIG).
options:
    resource_group:
        description:
            - The name of the resource group.
        required: true
        type: str
    name:
        description:
            - The name of the Shared Image Gallery.
            - Valid names consist of less than 80 alphanumeric characters, underscores and periods.
        required: true
        type: str
    location:
        description:
            - Resource location.
        type: str
    description:
        description:
            - The description of this Shared Image Gallery resource. This property is updatable.
        type: str
    state:
        description:
            - Assert the state of the Gallery.
            - Use C(present) to create or update an Gallery and C(absent) to delete it.
        default: present
        type: str
        choices:
            - absent
            - present
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags
author:
    - Zim Kalinowski (@zikalino)

'''

EXAMPLES = '''
- name: Create or update a simple gallery.
  azure_rm_gallery:
    resource_group: myResourceGroup
    name: myGallery1283
    location: West US
    description: This is the gallery description.
'''

RETURN = '''
id:
    description:
        - Resource ID.
    returned: always
    type: str
    sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/myResourceGroup/providers/Microsoft.Compute/galleries/myGallery1283"
'''

import time
import json
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_rest import GenericRestClient


class Actions:
    NoAction, Create, Update, Delete = range(4)


class AzureRMGalleries(AzureRMModuleBaseExt):
    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(
                type='str',
                required=True
            ),
            name=dict(
                type='str',
                required=True
            ),
            location=dict(
                type='str'
            ),
            description=dict(
                type='str',
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'absent']
            )
        )

        self.resource_group = None
        self.name = None
        self.gallery = None

        self.results = dict(changed=False)
        self.mgmt_client = None
        self.state = None
        self.url = None
        self.status_code = [200, 201, 202]
        self.to_do = Actions.NoAction

        self.body = {}
        self.body['properties'] = {}
        self.query_parameters = {}
        self.query_parameters['api-version'] = '2019-07-01'
        self.header_parameters = {}
        self.header_parameters['Content-Type'] = 'application/json; charset=utf-8'

        super(AzureRMGalleries, self).__init__(derived_arg_spec=self.module_arg_spec,
                                               supports_check_mode=True,
                                               supports_tags=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()) + ['tags']:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            elif kwargs[key] is not None:
                if key == 'description':
                    self.body['properties']['description'] = kwargs[key]
                else:
                    self.body[key] = kwargs[key]

        old_response = None
        response = None

        self.mgmt_client = self.get_mgmt_svc_client(GenericRestClient,
                                                    base_url=self._cloud_environment.endpoints.resource_manager)

        resource_group = self.get_resource_group(self.resource_group)

        if 'location' not in self.body:
            self.body['location'] = resource_group.location

        self.url = ('/subscriptions' +
                    '/{{ subscription_id }}' +
                    '/resourceGroups' +
                    '/{{ resource_group }}' +
                    '/providers' +
                    '/Microsoft.Compute' +
                    '/galleries' +
                    '/{{ gallery_name }}')
        self.url = self.url.replace('{{ subscription_id }}', self.subscription_id)
        self.url = self.url.replace('{{ resource_group }}', self.resource_group)
        self.url = self.url.replace('{{ gallery_name }}', self.name)

        old_response = self.get_resource()

        if not old_response:
            self.log("Gallery instance doesn't exist")

            if self.state == 'absent':
                self.log("Old instance didn't exist")
            else:
                self.to_do = Actions.Create
        else:
            self.log('Gallery instance already exists')

            if self.state == 'absent':
                self.to_do = Actions.Delete
            else:
                if self.body['properties'].get('description') is not None and\
                   self.body['properties']['description'] != old_response['properties'].get('description'):
                    self.to_do = Actions.Update
                else:
                    self.body['properties']['description'] = old_response['properties'].get('description')

                update_tags, new_tags = self.update_tags(old_response.get('tags'))
                if update_tags:
                    self.to_do = Actions.Update
                    self.body['tags'] = new_tags

        if self.to_do == Actions.Create:
            self.log('Need to Create the Gallery instance')
            if self.check_mode:
                self.results['changed'] = True
                return self.results
            response = self.create_resource()
            # if not old_response:
            self.results['changed'] = True
            # else:
            #     self.results['changed'] = old_response.__ne__(response)
            self.log('Creation done')
        elif self.to_do == Actions.Update:
            self.log('Need to Update the Gallery instance')
            if self.check_mode:
                self.results['changed'] = True
                return self.results
            response = self.update_resource()
            # if not old_response:
            self.results['changed'] = True
            self.log('Update done')
        elif self.to_do == Actions.Delete:
            self.log('Gallery instance deleted')
            self.results['changed'] = True

            if self.check_mode:
                return self.results

            self.delete_resource()

            # make sure instance is actually deleted, for some Azure resources, instance is hanging around
            # for some time after deletion -- this should be really fixed in Azure
            while self.get_resource():
                time.sleep(20)
        else:
            self.log('Gallery instance unchanged')
            self.results['changed'] = False
            response = old_response

        if response:
            self.results["id"] = response["id"]

        return self.results

    def update_resource(self):
        # self.log('Updating the Gallery instance {0}'.format(self.))

        try:
            response = self.mgmt_client.query(self.url,
                                              'PATCH',
                                              self.query_parameters,
                                              self.header_parameters,
                                              self.body,
                                              self.status_code,
                                              600,
                                              30)
        except Exception as exc:
            self.log('Error attempting to update the Gallery instance.')
            self.fail('Error updating the Gallery instance: {0}'.format(str(exc)))

        if hasattr(response, 'body'):
            response = json.loads(response.body())
        elif hasattr(response, 'context'):
            response = response.context['deserialized_data']
        else:
            self.fail("Updating fail, no match message return, return info as {0}".format(response))

        return response

    def create_resource(self):
        # self.log('Creating the Gallery instance {0}'.format(self.))

        try:
            response = self.mgmt_client.query(self.url,
                                              'PUT',
                                              self.query_parameters,
                                              self.header_parameters,
                                              self.body,
                                              self.status_code,
                                              600,
                                              30)
        except Exception as exc:
            self.log('Error attempting to create the Gallery instance.')
            self.fail('Error creating the Gallery instance: {0}'.format(str(exc)))

        if hasattr(response, 'body'):
            response = json.loads(response.body())
        elif hasattr(response, 'context'):
            response = response.context['deserialized_data']
        else:
            self.fail("Create fail, no match message return, return info as {0}".format(response))

        return response

    def delete_resource(self):
        # self.log('Deleting the Gallery instance {0}'.format(self.))
        try:
            response = self.mgmt_client.query(self.url,
                                              'DELETE',
                                              self.query_parameters,
                                              self.header_parameters,
                                              None,
                                              self.status_code,
                                              600,
                                              30)
        except Exception as e:
            self.log('Error attempting to delete the Gallery instance.')
            self.fail('Error deleting the Gallery instance: {0}'.format(str(e)))

        return True

    def get_resource(self):
        # self.log('Checking if the Gallery instance {0} is present'.format(self.))
        found = False
        try:
            response = self.mgmt_client.query(self.url,
                                              'GET',
                                              self.query_parameters,
                                              self.header_parameters,
                                              None,
                                              self.status_code,
                                              600,
                                              30)
            response = json.loads(response.body())
            found = True
            self.log("Response : {0}".format(response))
            # self.log("AzureFirewall instance : {0} found".format(response.name))
        except Exception as e:
            self.log('Did not find the AzureFirewall instance.')
        if found is True:
            return response

        return False


def main():
    AzureRMGalleries()


if __name__ == '__main__':
    main()



# Copyright (c) 2023, jantari (https://github.com/jantari)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: bitwarden_secrets_manager
author:
  - jantari (@jantari)
requirements:
  - bws (command line utility)
short_description: Retrieve secrets from Bitwarden Secrets Manager
version_added: 7.2.0
description:
  - Retrieve secrets from Bitwarden Secrets Manager.
options:
  _terms:
    description: Secret ID(s) to fetch values for.
    required: true
    type: list
    elements: str
  bws_access_token:
    description: The BWS access token to use for this lookup.
    env:
      - name: BWS_ACCESS_TOKEN
    required: true
    type: str
"""

EXAMPLES = r"""
- name: Get a secret relying on the BWS_ACCESS_TOKEN environment variable for authentication
  ansible.builtin.debug:
    msg: >-
      {{ lookup("community.general.bitwarden_secrets_manager", "2bc23e48-4932-40de-a047-5524b7ddc972") }}

- name: Get a secret passing an explicit access token for authentication
  ansible.builtin.debug:
    msg: >-
      {{
        lookup(
          "community.general.bitwarden_secrets_manager",
          "2bc23e48-4932-40de-a047-5524b7ddc972",
          bws_access_token="9.4f570d14-4b54-42f5-bc07-60f4450b1db5.YmluYXJ5LXNvbWV0aGluZy0xMjMK:d2h5IGhlbGxvIHRoZXJlCg=="
        )
      }}

- name: Get two different secrets each using a different access token for authentication
  ansible.builtin.debug:
    msg:
      - '{{ lookup("community.general.bitwarden_secrets_manager", "2bc23e48-4932-40de-a047-5524b7ddc972", bws_access_token=token1) }}'
      - '{{ lookup("community.general.bitwarden_secrets_manager", "9d89af4c-eb5d-41f5-bb0f-4ae81215c768", bws_access_token=token2) }}'
  vars:
    token1: "9.4f570d14-4b54-42f5-bc07-60f4450b1db5.YmluYXJ5LXNvbWV0aGluZy0xMjMK:d2h5IGhlbGxvIHRoZXJlCg=="
    token2: "1.69b72797-6ea9-4687-a11e-848e41a30ae6.YW5zaWJsZSBpcyBncmVhdD8K:YW5zaWJsZSBpcyBncmVhdAo="

- name: Get just the value of a secret
  ansible.builtin.debug:
    msg: >-
      {{ lookup("community.general.bitwarden_secrets_manager", "2bc23e48-4932-40de-a047-5524b7ddc972").value }}
"""

RETURN = r"""
_raw:
  description: List containing one or more secrets.
  type: list
  elements: dict
"""

from subprocess import PIPE, Popen
from time import sleep

from ansible.errors import AnsibleLookupError
from ansible.module_utils.common.text.converters import to_text
from ansible.parsing.ajson import AnsibleJSONDecoder
from ansible.plugins.lookup import LookupBase

from ansible_collections.community.general.plugins.module_utils._version import LooseVersion
from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms


class BitwardenSecretsManagerException(AnsibleLookupError):
    pass


class BitwardenSecretsManager:
    def __init__(self, path="bws"):
        self._cli_path = path
        self._max_retries = 3
        self._retry_delay = 1

    @property
    def cli_path(self):
        return self._cli_path

    def _run_with_retry(self, args, stdin=None, retries=0):
        out, err, rc = self._run(args, stdin)

        if rc != 0:
            if retries >= self._max_retries:
                raise BitwardenSecretsManagerException("Max retries exceeded. Unable to retrieve secret.")

            if "Too many requests" in err:
                delay = self._retry_delay * (2**retries)
                sleep(delay)
                return self._run_with_retry(args, stdin, retries + 1)
            else:
                raise BitwardenSecretsManagerException(f"Command failed with return code {rc}: {err}")

        return out, err, rc

    def _run(self, args, stdin=None):
        p = Popen([self.cli_path] + args, stdout=PIPE, stderr=PIPE, stdin=PIPE)
        out, err = p.communicate(stdin)
        rc = p.wait()
        return to_text(out, errors="surrogate_or_strict"), to_text(err, errors="surrogate_or_strict"), rc

    def get_bws_version(self):
        """Get the version of the Bitwarden Secrets Manager CLI."""
        out, err, rc = self._run(["--version"])
        if rc != 0:
            raise BitwardenSecretsManagerException(to_text(err))
        # strip the prefix and grab the last segment, the version number
        return out.split()[-1]

    def get_secret(self, secret_id, bws_access_token):
        """Get and return the secret with the given secret_id."""

        # Prepare set of params for Bitwarden Secrets Manager CLI
        # Color output was not always disabled correctly with the default 'auto' setting so explicitly disable it.
        params = ["--color", "no", "--access-token", bws_access_token]

        # bws version 0.3.0 introduced a breaking change in the command line syntax:
        # pre-0.3.0: verb noun
        # 0.3.0 and later: noun verb
        bws_version = self.get_bws_version()
        if LooseVersion(bws_version) < LooseVersion("0.3.0"):
            params.extend(["get", "secret", secret_id])
        else:
            params.extend(["secret", "get", secret_id])

        out, err, rc = self._run_with_retry(params)
        if rc != 0:
            raise BitwardenSecretsManagerException(to_text(err))

        return AnsibleJSONDecoder().raw_decode(out)[0]


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs):
        self.set_options(var_options=variables, direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)
        bws_access_token = self.get_option("bws_access_token")

        return [_bitwarden_secrets_manager.get_secret(term, bws_access_token) for term in terms]


_bitwarden_secrets_manager = BitwardenSecretsManager()



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_ml_schedule
version_added: "3.19.0"
short_description: Create, Disable, Enable or Delete an Azure Machine Learning Schedule
description:
    - Create, Disable, Enable or Delete an Azure Machine Learning Schedule.
options:
    name:
        description:
            - Name of the Azure ML schedule.
        type: str
        required: false
    resource_group:
        description:
            - Name of resource group.
        required: true
        type: str
    ml_workspace:
        description:
            - Name of the Azure ML workspace.
        required: true
        type: str
    resource_definition:
        description:
            - "Azure ML schedule specification. The YAML reference docs
              for schedule can be found at:
              https://aka.ms/ml-cli-v2-schedule-yaml-reference."
        type: str
        required: false
    state:
        description:
            - State of the Schedule. Use C(present) to create
              or update. C(disabled) to disable, C(enabled) to
              enable.
        default: present
        type: str
        choices:
            - present
            - disabled
            - enabled
            - absent

extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags

author:
    - Bill Peck (@p3ck)
'''

EXAMPLES = '''
- name: Create ML Schedule
  azure.azcollection.azure_rm_ml_schedule:
    name: schedulexxxxxxxxxx
    resource_group: xxxxx-ml-schedule
    ml_workspace: workspace-xxxxxxxxxx
    resource_definition: |
      $schema: https://azuremlschemas.azureedge.net/latest/schedule.schema.json
      name: simple_cron_job_schedule
      display_name: Simple cron job schedule
      description: a simple hourly cron job schedule

      trigger:
        type: cron
        expression: "1 * * * *"
        start_time: "2022-07-10T10:00:00" # optional
        time_zone: "Pacific Standard Time" # optional

      # create_job: azureml:simple-pipeline-job
      create_job: ./targets/azure_rm_ml_schedule/files/simple-pipeline-job.yml
'''

RETURN = '''
changed:
    description:
        - Whether the resource is changed.
    returned: always
    type: bool
    sample: false
ml_schedule:
    description:
        - Schedule that was just created or updated.
    returned: always
    type: dict
    sample: {
      "create_job": {
          "compute": "azureml:cpu-cluster",
          "display_name": "hello_pipeline_abc",
          "experiment_name": "Default",
          "inputs": {
              "hello_string_top_level_input": "hello world"
          },
          "jobs": {
              "a": {
                  "component": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-schedule/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/components/azureml_anonymous/versions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                  "inputs": {
                      "hello_string": {
                          "path": "${{ '{{' }}parent.inputs.hello_string_top_level_input{{ '}}' }}"
                      }
                  },
                  "type": "command"
              },
              "b": {
                  "component": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-schedule/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/components/azureml_anonymous/versions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                  "type": "command"
              },
              "c": {
                  "component": "azureml:/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/resourceGroups/xxxxx-ml-schedule/providers/Microsoft.MachineLearningServices/workspaces/workspace-xxxxxxxxxx/components/azureml_anonymous/versions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
                  "inputs": {
                      "world_input": {
                          "path": "${{ '{{' }}parent.jobs.b.outputs.world_output{{ '}}' }}"
                      }
                  },
                  "type": "command"
              }
          },
          "status": "NotStarted",
          "type": "pipeline"
      },
      "creation_context": {
          "created_at": "2026-05-14T16:46:52.141013+00:00",
          "created_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
          "created_by_type": "User",
          "last_modified_at": "2026-05-14T16:47:49.168130+00:00",
          "last_modified_by": "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx",
          "last_modified_by_type": "User"
      },
      "description": "a simple hourly cron job schedule",
      "display_name": "Simple cron job schedule",
      "is_enabled": false,
      "name": "schedulexxxxxxxxxx",
      "provisioning_state": "Succeeded",
      "trigger": {
          "expression": "1 * * * *",
          "start_time": "2022-07-10 10:00:00",
          "time_zone": "Pacific Standard Time",
          "type": "cron"
      }
    }
'''  # NOQA


try:
    import io
    from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ml import MLClientCommon
    from azure.core.exceptions import ResourceNotFoundError
    from azure.ai.ml.entities._load_functions import load_schedule
    from azure.ai.ml.entities._load_functions import _try_load_yaml_dict
    from azure.ai.ml.exceptions import ValidationException
except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMMLSchedule(MLClientCommon):

    PARAM_OVERRIDES = ['name',
                       'tags',
                       ]

    def __init__(self):

        self.module_arg_spec = dict(
            name=dict(
                type='str',
                required=False,
            ),
            resource_definition=dict(
                type='str',
                required=False,
            ),
            resource_group=dict(
                type='str',
                required=True,
            ),
            ml_workspace=dict(
                type='str',
                required=True,
            ),
            state=dict(
                type='str',
                default='present',
                choices=['present', 'disabled', 'enabled', 'absent'],
            ),
        )

        self._client = None
        self.ml_registry = None
        self.state = None

        self.results = dict(
            changed=False,
            compare=[],
            ml_schedule={}
        )

        required_if = [
            ('state', 'present', ['resource_definition']),
            ('resource_definition', None, ['name'])
        ]

        super(AzureRMMLSchedule, self).__init__(self.module_arg_spec,
                                                supports_check_mode=False,
                                                required_if=required_if,
                                                )

    def exec_module(self, **kwargs):

        for key in list(self.module_arg_spec.keys()) + ['tags']:
            setattr(self, key, kwargs[key])

        changed = False

        # Process resource_file yaml and passed in parameters
        try:
            resource_definition = io.StringIO(self.resource_definition)
            yaml_dict = _try_load_yaml_dict(resource_definition)
        except ValidationException:
            yaml_dict = dict()
        schedule_name = self.name or yaml_dict.get('name', None)

        ml_schedule_info = self.get(name=schedule_name)

        if self.state == 'present':
            params_override = self.update_params(kwargs, self.PARAM_OVERRIDES)

            resource_definition = io.StringIO(self.resource_definition)
            ml_schedule = load_schedule(source=resource_definition,
                                        params_override=params_override)

            if ml_schedule_info:
                # Update
                orig_schedule_info = self.entity_to_dict(ml_schedule_info)

                response = self.client.begin_create_or_update(ml_schedule)
                ml_schedule_object = self.get_poller_result(response)
                ml_schedule_info = self.entity_to_dict(ml_schedule_object)

                changed = not self.default_compare(
                    {},
                    ml_schedule_info,
                    orig_schedule_info,
                    '',
                    self.results
                )
            else:
                # Create
                changed = True
                response = self.client.begin_create_or_update(ml_schedule)
                ml_schedule_object = self.get_poller_result(response)
                ml_schedule_info = self.entity_to_dict(ml_schedule_object)
        elif self.state == 'absent':
            # Delete
            if ml_schedule_info:
                changed = True
                response = self.client.schedules.begin_delete(name=schedule_name)
                result = self.get_poller_result(response)
                ml_schedule_info = self.get(schedule_name, as_dict=True)
        elif self.state == 'enabled':
            # Enable
            list_view_type = self.get_schedule_list_view_type('disabled')
            results = self.client.schedules.list(list_view_type=list_view_type)
            ml_schedules = [x.name for x in results]

            if schedule_name in ml_schedules:
                changed = True
                self.enable(name=schedule_name)
            ml_schedule_info = self.get(name=schedule_name,
                                        as_dict=True)
        elif self.state == 'disabled':
            # Disable
            list_view_type = self.get_schedule_list_view_type('enabled')
            results = self.client.schedules.list(list_view_type=list_view_type)
            ml_schedules = [x.name for x in results]

            if schedule_name in ml_schedules:
                changed = True
                self.disable(name=schedule_name)
            ml_schedule_info = self.get(name=schedule_name,
                                        as_dict=True)

        self.results['ml_schedule'] = ml_schedule_info
        self.results['changed'] = changed
        return self.results

    def update_params(self, kwargs, param_overrides, ml_schedule=None):
        params_override = []

        # If ml_schedule is defined use those values as a base.
        # Update can't change some of these values but we need
        # to keep the same values as before the update.
        if ml_schedule:
            items = self.entity_to_dict(ml_schedule).items()
            params_override = [{k: v} for k, v in items]

        # Update values based on what param_overrides says we can update.
        for key in list(self.module_arg_spec.keys()):
            if kwargs[key] is not None and key in param_overrides:
                params_override.append({key: kwargs[key]})
        return params_override

    def get(self, name, as_dict=False):
        try:
            # Attempt to retrieve ml_schedule based on name
            ml_schedule_info = self.client.schedules.get(name=name)
            if as_dict:
                ml_schedule_info = self.entity_to_dict(ml_schedule_info)
        except ResourceNotFoundError:
            ml_schedule_info = None
        return ml_schedule_info

    def enable(self, name=None):
        try:
            response = self.client.schedules.begin_enable(name=name)
            ml_schedule_object = self.get_poller_result(response)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.fail("Error enabling the schedule instance: {0}".format(str(e)))

    def disable(self, name=None):
        try:
            response = self.client.schedules.begin_disable(name=name)
            ml_schedule_object = self.get_poller_result(response)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.fail("Error disabling the schedule instance: {0}".format(str(e)))


def main():
    AzureRMMLSchedule()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2023 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
The module file for ios_evpn_evi
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: ios_evpn_evi
short_description: Resource module to configure L2VPN EVPN EVI.
description: This module provides declarative management of L2VPN EVPN EVI on Cisco IOS network
  devices.
version_added: 5.3.0
author: Padmini Priyadarshini Sivaraj (@PadminiSivaraj)
notes:
  - Tested against Cisco IOS-XE device with Version 17.13.01 on Cat9k on CML.
  - This module works with connection C(network_cli).
    See U(https://docs.ansible.com/ansible/latest/network/user_guide/platform_ios.html)
options:
  config:
    description: A dictionary of L2VPN Ethernet Virtual Private Network (EVPN) EVI configuration
    type: list
    elements: dict
    suboptions:
      evi:
        description: EVPN instance value
        type: int
        required: True
      default_gateway:
        description: Default Gateway parameters
        type: dict
        suboptions:
          advertise:
            description: Advertise Default Gateway MAC/IP routes
            type: dict
            suboptions:
              enable:
                description: Enable advertisement of Default Gateway MAC/IP routes
                type: bool
              disable:
                description: Disable advertisement of Default Gateway MAC/IP routes
                type: bool
      ip:
        description: IP parameters
        type: dict
        suboptions:
          local_learning:
            description: IP local learning
            type: dict
            suboptions:
              enable:
                description: Enable IP local learning
                type: bool
              disable:
                description: Disable IP local learning
                type: bool
      encapsulation:
        description: EVPN encapsulation type
        type: str
        choices:
        - vxlan
        default: vxlan
      replication_type:
        description: Method for replicating BUM traffic
        type: str
        choices:
        - ingress
        - static
      route_distinguisher:
        description: EVPN Route Distinguisher
        type: str
  running_config:
    description:
      - This option is used only with state I(parsed).
      - The value of this option should be the output received from the IOS
        device by executing the command B(sh running-config nve | section ^l2vpn evpn$).
      - The state I(parsed) reads the configuration from C(running_config)
        option and transforms it into Ansible structured data as per the
        resource module's argspec and the value is then returned in the
        I(parsed) key within the result.
    type: str
  state:
    description:
      - The state the configuration should be left in
    type: str
    choices:
      - merged
      - replaced
      - overridden
      - deleted
      - gathered
      - rendered
      - parsed
    default: merged
"""

EXAMPLES = """

# Using state merged

# Before state:
# -------------
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# !
# l2vpn evpn instance 201 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type ingress

# - name: Merge provided configuration with device configuration
#   cisco.ios.ios_evpn_evi:
#     config:
#       - evi: 101
#         replication_type: ingress
#         route_distinguisher: '1:1'
#         default_gateway:
#           advertise:
#             enable: False
#         ip:
#           local_learning:
#             enable: True
#
#       - evi: 202
#         replication_type: static
#         default_gateway:
#           advertise:
#             enable: True
#         ip:
#           local_learning:
#             disable: True
#     state: merged

# Commands Fired:
# ---------------
# "commands": [
#     "l2vpn evpn instance 101 vlan-based",
#     "ip local-learning enable",
#     "replication-type ingress",
#     "rd 1:1",
#     "l2vpn evpn instance 202 vlan-based",
#     "default-gateway advertise enable",
#     "ip local-learning disable",
#     "replication-type static"
#     ],

# After state:
# ------------
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  rd 1:1
#  replication-type ingress
#  ip local-learning enable
# !
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# !
# l2vpn evpn instance 201 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type static
#  ip local-learning disable
#  default-gateway advertise enable


# Using state replaced

# Before state:
# -------------
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  rd 1:1
#  replication-type ingress
#  ip local-learning enable
# !
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# !
# l2vpn evpn instance 201 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type static
#  ip local-learning disable
#  default-gateway advertise enable

# - name: Replaces the device configuration with the provided configuration
#   cisco.ios.ios_evpn_evi:
#     config:
#       - evi: 101
#         replication_type: ingress
#         default_gateway:
#           advertise:
#             enable: True
#       - evi: 202
#         replication_type: ingress
#     state: replaced

# Commands Fired:
# ---------------
# "commands": [
#     "l2vpn evpn instance 101 vlan-based",
#     "default-gateway advertise enable",
#     "no ip local-learning enable",
#     "no rd 1:1",
#     "l2vpn evpn instance 202 vlan-based",
#     "no default-gateway advertise enable",
#     "no ip local-learning disable",
#     "replication-type ingress"
#     ],

# After state:
# ------------
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  replication-type ingress
#  default-gateway advertise enable
# !
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# !
# l2vpn evpn instance 201 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type ingress

# Using state overridden

# Before state:
# -------------
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  replication-type ingress
#  default-gateway advertise enable
# !
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# !
# l2vpn evpn instance 201 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type ingress

# - name: Override the device configuration with provided configuration
#   cisco.ios.ios_evpn_evi:
#     config:
#       - evi: 101
#         replication_type: ingress
#         default_gateway:
#           advertise:
#             enable: True
#       - evi: 202
#         replication_type: static
#         default_gateway:
#           advertise:
#             enable: True
#     state: overridden

# Commands Fired:
# ---------------
# "commands": [
#     "no l2vpn evpn instance 102 vlan-based",
#     "no l2vpn evpn instance 201 vlan-based",
#     "l2vpn evpn instance 202 vlan-based",
#     "default-gateway advertise enable",
#     "replication-type static"
#     ],

# After state:
# ------------
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  replication-type ingress
#  default-gateway advertise enable
# !
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type static
#  default-gateway advertise enable


# Using state Deleted

# Before state:
# -------------
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  replication-type ingress
#  default-gateway advertise enable
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type static
#  default-gateway advertise enable

# - name: "Delete the given EVI(s)"
#   cisco.ios.ios_evpn_evi:
#     config:
#       - evi: 101
#     state: deleted

# Commands Fired:
# ---------------
# "commands": [
#       "no l2vpn evpn instance 101 vlan-based"
#       ],

# After state:
# -------------
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type static
#  default-gateway advertise enable

# Using state Deleted without any config passed

# Before state:
# -------------
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type static
#  default-gateway advertise enable

# - name: "Delete ALL EVIs"
#   cisco.ios.ios_evpn_evi:
#     state: deleted

# Commands Fired:
# ---------------
# "commands": [
#     "no l2vpn evpn instance 102 vlan-based",
#     "no l2vpn evpn instance 202 vlan-based"
#     ],

# After state:
# -------------
# !

# Using gathered

# Before state:
# -------------
#
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# !
# l2vpn evpn instance 201 vlan-based
#  encapsulation vxlan
#  replication-type static
# !
# l2vpn evpn instance 202 vlan-based
#  encapsulation vxlan
#  replication-type ingress

# - name: Gather facts for evpn_evi
#   cisco.ios.ios_evpn_evi:
#     config:
#     state: gathered

# Task Output:
# ------------
#
# gathered:
#   - evi: 101
#     encapsulation: vxlan
#     replication_type: static
#   - evi: 102
#     encapsulation: vxlan
#     replication_type: ingress
#   - evi: 201
#     encapsulation: vxlan
#     replication_type: static
#   - evi: 202
#     encapsulation: vxlan
#     replication_type: ingress

# Using Rendered

# - name: Rendered the provided configuration with the existing running configuration
#   cisco.ios.ios_evpn_evi:
#     config:
#       - evi: 101
#         replication_type: ingress
#         default_gateway:
#           advertise:
#             enable: True
#       - evi: 202
#         replication_type: ingress
#     state: rendered

# Task Output:
# ------------
#
# rendered:
# - l2vpn evpn instance 101 vlan-based
# - default-gateway advertise enable
# - replication-type ingress
# - l2vpn evpn instance 202 vlan-based
# - replication-type ingress


# Using parsed

# File: parsed.cfg
# ----------------
#
# l2vpn evpn instance 101 vlan-based
#  encapsulation vxlan
#  replication-type ingress
#  default-gateway advertise enable
# !
# l2vpn evpn instance 102 vlan-based
#  encapsulation vxlan
#  replication-type ingress
# !

# - name: Parse the commands for provided configuration
#   cisco.ios.ios_evpn_evi:
#     running_config: "{{ lookup('file', 'parsed.cfg') }}"
#     state: parsed

# Task Output:
# ------------
#
# parsed:
#   - evi: 101
#     encapsulation: vxlan
#     replication_type: ingress
#     default_gateway:
#       advertise:
#         enable: true
#   - evi: 102
#     encapsulation: vxlan
#     replication_type: ingress
"""

RETURN = """
before:
  description: The configuration prior to the module execution.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted) or C(purged)
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
after:
  description: The resulting configuration after module execution.
  returned: when changed
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
commands:
  description: The set of commands pushed to the remote device.
  returned: when I(state) is C(merged), C(replaced), C(overridden), C(deleted) or C(purged)
  type: list
  sample:
    - "l2vpn evpn instance 101 vlan-based"
    - "encapsulation vxlan"
    - "replication-type ingress"
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.evpn_evi.evpn_evi import (
    Evpn_eviArgs,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.config.evpn_evi.evpn_evi import (
    Evpn_evi,
)


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    module = AnsibleModule(
        argument_spec=Evpn_eviArgs.argument_spec,
        mutually_exclusive=[["config", "running_config"]],
        required_if=[
            ["state", "merged", ["config"]],
            ["state", "replaced", ["config"]],
            ["state", "overridden", ["config"]],
            ["state", "rendered", ["config"]],
            ["state", "parsed", ["running_config"]],
        ],
        supports_check_mode=True,
    )

    result = Evpn_evi(module).execute_module()
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2018 Felix Fontein (@felixfontein)
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: acme_inspect
author: "Felix Fontein (@felixfontein)"
short_description: Send direct requests to an ACME server
description:
  - Allows to send direct requests to an ACME server with the L(ACME protocol,https://tools.ietf.org/html/rfc8555), which
    is supported by CAs such as L(Let's Encrypt,https://letsencrypt.org/).
  - This module can be used to debug failed certificate request attempts, for example when M(community.crypto.acme_certificate)
    fails or encounters a problem which you wish to investigate.
  - The module can also be used to directly access features of an ACME servers which are not yet supported by the Ansible
    ACME modules.
notes:
  - The O(account_uri) option must be specified for properly authenticated ACME v2 requests (except a C(new-account) request).
  - "Using the C(ansible) tool, M(community.crypto.acme_inspect) can be used to directly execute ACME requests without the
    need of writing a playbook. For example, the following command retrieves the ACME account with ID 1 from Let's Encrypt
    (assuming C(/path/to/key) is the correct private account key): C(ansible localhost -m acme_inspect -a \"account_key_src=/path/to/key
    acme_directory=https://acme-v02.api.letsencrypt.org/directory account_uri=https://acme-v02.api.letsencrypt.org/acme/acct/1
    method=get url=https://acme-v02.api.letsencrypt.org/acme/acct/1\")."
seealso:
  - name: Automatic Certificate Management Environment (ACME)
    description: The specification of the ACME protocol (RFC 8555).
    link: https://tools.ietf.org/html/rfc8555
  - name: ACME TLS ALPN Challenge Extension
    description: The specification of the C(tls-alpn-01) challenge (RFC 8737).
    link: https://www.rfc-editor.org/rfc/rfc8737.html
extends_documentation_fragment:
  - community.crypto._acme.basic
  - community.crypto._acme.account
  - community.crypto._attributes
  - community.crypto._attributes.actiongroup_acme
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none
  idempotent:
    support: none
options:
  url:
    description:
      - The URL to send the request to.
      - Must be specified if O(method) is not V(directory-only).
    type: str
  method:
    description:
      - The method to use to access the given URL on the ACME server.
      - The value V(post) executes an authenticated POST request. The content must be specified in the O(content) option.
      - The value V(get) executes an authenticated POST-as-GET request for ACME v2, and a regular GET request for ACME v1.
      - The value V(directory-only) only retrieves the directory, without doing a request.
    type: str
    default: get
    choices:
      - get
      - post
      - directory-only
  content:
    description:
      - An encoded JSON object which will be sent as the content if O(method) is V(post).
      - Required when O(method) is V(post), and not allowed otherwise.
    type: str
  fail_on_acme_error:
    description:
      - If O(method) is V(post) or V(get), make the module fail in case an ACME error is returned.
    type: bool
    default: true
"""

EXAMPLES = r"""
---
- name: Get directory
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    method: directory-only
  register: directory

- name: Create an account
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    url: "{{ directory.newAccount}}"
    method: post
    content: '{"termsOfServiceAgreed":true}'
  register: account_creation
  # account_creation.headers.location contains the account URI
  # if creation was successful

- name: Get account information
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    account_uri: "{{ account_creation.headers.location }}"
    url: "{{ account_creation.headers.location }}"
    method: get

- name: Update account contacts
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    account_uri: "{{ account_creation.headers.location }}"
    url: "{{ account_creation.headers.location }}"
    method: post
    content: '{{ account_info | to_json }}'
  vars:
    account_info:
      # For valid values, see
      # https://tools.ietf.org/html/rfc8555#section-7.3
      contact:
        - mailto:me@example.com

- name: Create certificate order
  community.crypto.acme_certificate:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    account_uri: "{{ account_creation.headers.location }}"
    csr: /etc/pki/cert/csr/sample.com.csr
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    challenge: http-01
    modify_account: false
  register: certificate_request

# Assume something went wrong. certificate_request.order_uri contains
# the order URI.

- name: Get order information
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    account_uri: "{{ account_creation.headers.location }}"
    url: "{{ certificate_request.order_uri }}"
    method: get
  register: order

- name: Get first authz for order
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    account_uri: "{{ account_creation.headers.location }}"
    url: "{{ order.output_json.authorizations[0] }}"
    method: get
  register: authz

- name: Get HTTP-01 challenge for authz
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    account_uri: "{{ account_creation.headers.location }}"
    url: "{{ authz.output_json.challenges | selectattr('type', 'equalto', 'http-01') }}"
    method: get
  register: http01challenge

- name: Activate HTTP-01 challenge manually
  community.crypto.acme_inspect:
    acme_directory: https://acme-staging-v02.api.letsencrypt.org/directory
    account_key_src: /etc/pki/cert/private/account.key
    account_uri: "{{ account_creation.headers.location }}"
    url: "{{ http01challenge.url }}"
    method: post
    content: '{}'
"""

RETURN = r"""
directory:
  description: The ACME directory's content.
  returned: always
  type: dict
  sample: {
    "a85k3x9f91A4": "https://community.letsencrypt.org/t/adding-random-entries-to-the-directory/33417",
    "keyChange": "https://acme-v02.api.letsencrypt.org/acme/key-change",
    "meta": {
      "caaIdentities": ["letsencrypt.org"],
      "termsOfService": "https://letsencrypt.org/documents/LE-SA-v1.2-November-15-2017.pdf",
      "website": "https://letsencrypt.org",
    },
    "newAccount": "https://acme-v02.api.letsencrypt.org/acme/new-acct",
    "newNonce": "https://acme-v02.api.letsencrypt.org/acme/new-nonce",
    "newOrder": "https://acme-v02.api.letsencrypt.org/acme/new-order",
    "revokeCert": "https://acme-v02.api.letsencrypt.org/acme/revoke-cert"
  }
headers:
  description: The request's HTTP headers (with lowercase keys).
  returned: always
  type: dict
  sample: {
    "boulder-requester": "12345",
    "cache-control": "max-age=0, no-cache, no-store",
    "connection": "close",
    "content-length": "904",
    "content-type": "application/json",
    "cookies": {},
    "cookies_string": "",
    "date": "Wed, 07 Nov 2018 12:34:56 GMT",
    "expires": "Wed, 07 Nov 2018 12:44:56 GMT",
    "link": '<https://letsencrypt.org/documents/LE-SA-v1.2-November-15-2017.pdf>;rel="terms-of-service"',
    "msg": "OK (904 bytes)",
    "pragma": "no-cache",
    "replay-nonce": "1234567890abcdefghijklmnopqrstuvwxyzABCDEFGH",
    "server": "nginx",
    "status": 200,
    "strict-transport-security": "max-age=604800",
    "url": "https://acme-v02.api.letsencrypt.org/acme/acct/46161",
    "x-frame-options": "DENY",
  }
output_text:
  description: The raw text output.
  returned: always
  type: str
  sample: "{\\n  \\\"id\\\": 12345,\\n  \\\"key\\\": {\\n    \\\"kty\\\": \\\"RSA\\\",\\n ..."
output_json:
  description: The output parsed as JSON.
  returned: if output can be parsed as JSON
  type: dict
  sample:
    - id: 12345
    - key:
        - kty: RSA
        - '...'
"""

import typing as t

from ansible.module_utils.common.text.converters import to_bytes, to_text

from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (
    ACMEClient,
    create_backend,
    create_default_argspec,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ACMEProtocolException,
    ModuleFailException,
)


def main() -> t.NoReturn:
    argument_spec = create_default_argspec(require_account_key=False)
    argument_spec.update_argspec(
        url={"type": "str"},
        method={
            "type": "str",
            "choices": ["get", "post", "directory-only"],
            "default": "get",
        },
        content={"type": "str"},
        fail_on_acme_error={"type": "bool", "default": True},
    )
    argument_spec.update(
        required_if=[
            ("method", "get", ["url"]),
            ("method", "post", ["url", "content"]),
            ("method", "get", ["account_key_src", "account_key_content"], True),
            ("method", "post", ["account_key_src", "account_key_content"], True),
        ],
    )
    module = argument_spec.create_ansible_module()
    backend = create_backend(module, needs_acme_v2=False)

    result: dict[str, t.Any] = {}
    changed = False
    try:
        # Get hold of ACMEClient and ACMEAccount objects (includes directory)
        client = ACMEClient(module=module, backend=backend)
        method: t.Literal["get", "post", "directory-only"] = module.params["method"]
        result["directory"] = client.directory.directory
        # Do we have to do more requests?
        if method != "directory-only":
            url = module.params["url"]
            fail_on_acme_error = module.params["fail_on_acme_error"]
            # Do request
            if method == "get":
                data, info = client.get_request(
                    url, parse_json_result=False, fail_on_error=False
                )
            elif method == "post":
                changed = True  # only POSTs can change
                data, info = client.send_signed_request(
                    url,
                    to_bytes(module.params["content"]),
                    parse_json_result=False,
                    encode_payload=False,
                    fail_on_error=False,
                )
            else:
                raise AssertionError("Can never be reached")  # pragma: no cover
            # Update results
            result.update(
                {
                    "headers": info,
                    "output_text": to_text(data),
                }
            )
            # See if we can parse the result as JSON
            try:
                result["output_json"] = module.from_json(to_text(data))
            except Exception:
                pass
            # Fail if error was returned
            if fail_on_acme_error and info["status"] >= 400:
                raise ACMEProtocolException(module=module, info=info, content=data)
        # Done!
        module.exit_json(changed=changed, **result)
    except ModuleFailException as e:
        e.do_fail(module=module, **result)


if __name__ == "__main__":
    main()



# Copyright (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: sesu
short_description: CA Privileged Access Manager
description:
  - This become plugins allows your remote/login user to execute commands as another user using the C(sesu) utility.
author: ansible (@nekonyuu)
options:
  become_user:
    description: User you 'become' to execute the task.
    type: string
    default: ''
    ini:
      - section: privilege_escalation
        key: become_user
      - section: sesu_become_plugin
        key: user
    vars:
      - name: ansible_become_user
      - name: ansible_sesu_user
    env:
      - name: ANSIBLE_BECOME_USER
      - name: ANSIBLE_SESU_USER
  become_exe:
    description: C(sesu) executable.
    type: string
    default: sesu
    ini:
      - section: privilege_escalation
        key: become_exe
      - section: sesu_become_plugin
        key: executable
    vars:
      - name: ansible_become_exe
      - name: ansible_sesu_exe
    env:
      - name: ANSIBLE_BECOME_EXE
      - name: ANSIBLE_SESU_EXE
  become_flags:
    description: Options to pass to C(sesu).
    type: string
    default: -H -S -n
    ini:
      - section: privilege_escalation
        key: become_flags
      - section: sesu_become_plugin
        key: flags
    vars:
      - name: ansible_become_flags
      - name: ansible_sesu_flags
    env:
      - name: ANSIBLE_BECOME_FLAGS
      - name: ANSIBLE_SESU_FLAGS
  become_pass:
    description: Password to pass to C(sesu).
    type: string
    required: false
    vars:
      - name: ansible_become_password
      - name: ansible_become_pass
      - name: ansible_sesu_pass
    env:
      - name: ANSIBLE_BECOME_PASS
      - name: ANSIBLE_SESU_PASS
    ini:
      - section: sesu_become_plugin
        key: password
"""

from ansible.plugins.become import BecomeBase


class BecomeModule(BecomeBase):
    name = "community.general.sesu"

    prompt = "Please enter your password:"
    fail = missing = ("Sorry, try again with sesu.",)

    def build_become_command(self, cmd, shell):
        super().build_become_command(cmd, shell)

        if not cmd:
            return cmd

        become = self.get_option("become_exe")

        flags = self.get_option("become_flags")
        user = self.get_option("become_user")
        return f"{become} {flags} {user} -c {self._build_success_command(cmd, shell)}"



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2016, Ansible, inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r'''
module: win_region
short_description: Set the region and format settings
description:
    - Set the location settings of a Windows Server.
    - Set the format settings of a Windows Server.
    - Set the unicode language settings of a Windows Server.
    - Copy across these settings to the default profile.
version_added: 2.6.0
options:
    location:
        description:
            - The location to set for the current user, see
              U(https://msdn.microsoft.com/en-us/library/dd374073.aspx)
              for a list of GeoIDs you can use and what location it relates to.
            - This needs to be set if C(format) or C(unicode_language) is not
              set.
        type: str
    format:
        description:
            - The language format to set for the current user, see
              U(https://msdn.microsoft.com/en-us/library/system.globalization.cultureinfo.aspx)
              for a list of culture names to use.
            - This needs to be set if C(location) or C(unicode_language) is not set.
        type: str
    unicode_language:
        description:
            - The unicode language format to set for all users, see
              U(https://learn.microsoft.com/en-us/windows-hardware/manufacture/desktop/default-input-locales-for-windows-language-packs)
              for a list of culture names to use.
            - This needs to be set if C(location) or C(format) is not set. After setting this
              value a reboot is required for it to take effect.
        type: str
    copy_settings:
        description:
            - This will copy the current format and location values to new user
              profiles and the welcome screen. This will only run if
              C(location), C(format) or C(unicode_language) has resulted in a
              change. If this process runs then it will always result in a
              change.
        type: bool
        default: no
seealso:
- module: ansible.windows.win_timezone
author:
- Jordan Borean (@jborean93)
'''

EXAMPLES = r'''
- name: Set the region format to English United States
  ansible.windows.win_region:
    format: en-US

- name: Set the region format to English Australia and copy settings to new profiles
  ansible.windows.win_region:
    format: en-AU
    copy_settings: true

- name: Set the location to United States
  ansible.windows.win_region:
    location: 244

# Reboot when region settings change
- name: Set the unicode language to English Great Britain, reboot if required
  ansible.windows.win_region:
    unicode_language: en-GB
  register: result

- ansible.windows.win_reboot:
  when: result.restart_required

# Reboot when format, location or unicode has changed
- name: Set format, location and unicode to English Australia and copy settings, reboot if required
  ansible.windows.win_region:
    location: 12
    format: en-AU
    unicode_language: en-AU
  register: result

- ansible.windows.win_reboot:
  when: result.restart_required
'''

RETURN = r'''
restart_required:
    description: Whether a reboot is required for the change to take effect.
    returned: success
    type: bool
    sample: true
'''
